#include <arpa/inet.h>      // htonl(), ntohl(), etc
#include <string.h>         // memcpy(), strlen(), etc
#include <assert.h>         // assert()
#include <stdlib.h>         // malloc()
#include <stdio.h>          // fprintf()
#include <limits.h>         // strtol()
#include <pthread.h>        // pthread stuff
#include <sys/types.h>      // dir stuff
#include <sys/stat.h>       // file size
#include <dirent.h>         // dir stuff
#include <zlib.h>           // compress()

// #include <unistd.h>

#include "daap.h"
#include "id3.h"
#include "mpeg.h"
#include "buff.h"
#include "hash.h"

extern daap_database kDefaultDatabase;
extern daap_adbs     kDefaultAdbs;
extern daap_apso     kDefaultApso;

/**
// a little conversion table for getting meaningful, human-readable
// translations of otherwise hard-to-understand data (eg, 0x000C)
//
 const data_type data_type_list[9] =
 {
 {0x0001, "byte",    "unsigned char"},
 {0x0003, "short",   "unsigned short"},
 {0x0005, "int",     "unsigned int"},
 {0x0007, "long",    "unsigned long long"},
 {0x0009, "string",  "char[]"},
 {0x000A, "date",    "unsigned int"},
 {0x000B, "version", "unsigned int"},
 {0x000C, "list",    "n/a"}
 };
**/

// eighty-five various data types, plus an extra one which should be automatically
// initialized to zero, so that we should be able to iterate through this list,
// and just check for zero, to find the end ... i think.
//
const daap_type daap_type_list[] =
{
{"miid", "dmap.itemid",                     DAAP_TYPE_INT},
{"mikd", "dmap.itemkind",                   DAAP_TYPE_BYTE},
{"minm", "dmap.itemname",                   DAAP_TYPE_STRING},
{"mper", "dmap.persistentid",               DAAP_TYPE_LONG},
{"mcon", "dmap.container",                  DAAP_TYPE_LIST},
{"mcti", "dmap.containeritemid",            DAAP_TYPE_INT},
{"mpco", "dmap.parentcontainerid",          DAAP_TYPE_INT},
{"mstt", "dmap.status",                     DAAP_TYPE_INT},
{"msts", "dmap.statusstring",               DAAP_TYPE_STRING},
{"mimc", "dmap.itemcount",                  DAAP_TYPE_INT},
{"mctc", "dmap.containercount",             DAAP_TYPE_INT},
{"mrco", "dmap.returnedcount",              DAAP_TYPE_INT},
{"mtco", "dmap.specifiedtotalcount",        DAAP_TYPE_INT},
{"mlcl", "dmap.listing",                    DAAP_TYPE_LIST},
{"mlit", "dmap.listingitem",                DAAP_TYPE_LIST},
{"mbcl", "dmap.bag",                        DAAP_TYPE_LIST},
{"mdcl", "dmap.dictionary",                 DAAP_TYPE_LIST},
{"msrv", "dmap.serverinforesponse",         DAAP_TYPE_LIST},
{"msau", "dmap.authenticationmethod",       DAAP_TYPE_BYTE},
{"mslr", "dmap.loginrequired",              DAAP_TYPE_BYTE},
{"mpro", "dmap.protocolversion",            DAAP_TYPE_VERSION},
{"msal", "dmap.supportsautologout",         DAAP_TYPE_BYTE},
{"msup", "dmap.supportsupdate",             DAAP_TYPE_BYTE},
{"mspi", "dmap.supportspersistentids",      DAAP_TYPE_BYTE},
{"msex", "dmap.supportsextensions",         DAAP_TYPE_BYTE},
{"msbr", "dmap.supportsbrowse",             DAAP_TYPE_BYTE},
{"msqy", "dmap.supportsquery",              DAAP_TYPE_BYTE},
{"msix", "dmap.supportsindex",              DAAP_TYPE_BYTE},
{"msrs", "dmap.supportsresolve",            DAAP_TYPE_BYTE},
{"mstm", "dmap.timeoutinterval",            DAAP_TYPE_INT},
{"msdc", "dmap.databasescount",             DAAP_TYPE_INT},
{"mlog", "dmap.loginresponse",              DAAP_TYPE_LIST},
{"mlid", "dmap.sessionid",                  DAAP_TYPE_INT},
{"mupd", "dmap.updateresponse",             DAAP_TYPE_LIST},
{"musr", "dmap.serverrevision",             DAAP_TYPE_INT},
{"muty", "dmap.updatetype",                 DAAP_TYPE_BYTE},
{"mudl", "dmap.deletedidlisting",           DAAP_TYPE_LIST},
{"mccr", "dmap.contentcodesresponse",       DAAP_TYPE_LIST},
{"mcnm", "dmap.contentcodesnumber",         DAAP_TYPE_INT},
{"mcna", "dmap.contentcodesname",           DAAP_TYPE_STRING},
{"mcty", "dmap.contentcodestype",           DAAP_TYPE_SHORT},
{"apro", "daap.protocolversion",            DAAP_TYPE_VERSION},
{"avdb", "daap.serverdatabases",            DAAP_TYPE_LIST},
{"abro", "daap.databasebrowse",             DAAP_TYPE_LIST},
{"abal", "daap.browsealbumlisting",         DAAP_TYPE_LIST},
{"abar", "daap.browseartistlisting",        DAAP_TYPE_LIST},
{"abcp", "daap.browsecomposerlisting",      DAAP_TYPE_LIST},
{"abgn", "daap.browsegenrelisting",         DAAP_TYPE_LIST},
{"adbs", "daap.databasesongs",              DAAP_TYPE_LIST},
{"asal", "daap.songalbum",                  DAAP_TYPE_STRING},
{"agrp", "daap.songgrouping",               DAAP_TYPE_STRING},
{"asar", "daap.songartist",                 DAAP_TYPE_STRING},
{"asbt", "daap.songbeatsperminute",         DAAP_TYPE_SHORT},
{"asbr", "daap.songbitrate",                DAAP_TYPE_SHORT},
{"ascm", "daap.songcomment",                DAAP_TYPE_STRING},
{"asco", "daap.songcompilation",            DAAP_TYPE_BYTE},
{"ascp", "daap.songcomposer",               DAAP_TYPE_STRING},
{"asda", "daap.songdateadded",              DAAP_TYPE_DATE},
{"asdm", "daap.songdatemodified",           DAAP_TYPE_DATE},
{"asdc", "daap.songdisccount",              DAAP_TYPE_SHORT},
{"asdn", "daap.songdiscnumber",             DAAP_TYPE_SHORT},
{"asdb", "daap.songdisabled",               DAAP_TYPE_BYTE},
{"aseq", "daap.songeqpreset",               DAAP_TYPE_STRING},
{"asfm", "daap.songformat",                 DAAP_TYPE_STRING},
{"asgn", "daap.songgenre",                  DAAP_TYPE_STRING},
{"asdt", "daap.songdescription",            DAAP_TYPE_STRING},
{"asrv", "daap.songrelativevolume",         DAAP_TYPE_BYTE},
{"assr", "daap.songsamplerate",             DAAP_TYPE_INT},
{"assz", "daap.songsize",                   DAAP_TYPE_INT},
{"asst", "daap.songstarttime",              DAAP_TYPE_INT},
{"assp", "daap.songstoptime",               DAAP_TYPE_INT},
{"astm", "daap.songtime",                   DAAP_TYPE_INT},
{"astc", "daap.songtrackcount",             DAAP_TYPE_SHORT},
{"astn", "daap.songtracknumber",            DAAP_TYPE_SHORT},
{"asur", "daap.songuserrating",             DAAP_TYPE_BYTE},
{"asyr", "daap.songyear",                   DAAP_TYPE_SHORT},
{"asdk", "daap.songdatakind",               DAAP_TYPE_BYTE},
{"asul", "daap.songdataurl",                DAAP_TYPE_STRING},
{"aply", "daap.databaseplaylists",          DAAP_TYPE_LIST},
{"abpl", "daap.baseplaylist",               DAAP_TYPE_BYTE},
{"apso", "daap.playlistsongs",              DAAP_TYPE_LIST},
{"arsv", "daap.resolve",                    DAAP_TYPE_LIST},
{"arif", "daap.resolveinfo",                DAAP_TYPE_LIST},
{"aeNV", "com.apple.itunes.norm-volume",    DAAP_TYPE_INT},
{"aeSP", "com.apple.itunes.smart-playlist", DAAP_TYPE_BYTE}
};

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 01 [msrv] /server-info
#endif

daap_msrv *
msrv_Create ()
{
  daap_msrv * msrv;
  
  msrv = (daap_msrv *)malloc(sizeof(daap_msrv));
  assert(msrv != NULL);

  return msrv;
}

void
msrv_Init (daap_msrv * msrv)
{
  daap_int leng = 0;
  
  assert(msrv != NULL);
  
  bzero(msrv, sizeof(daap_msrv));

  leng += mstt_Init(&msrv->mstt, 200);      // dmap.status
  leng += mpro_Init(&msrv->mpro, 2, 0);     // dmap.protocolversion
  leng += apro_Init(&msrv->apro, 2, 0);     // dmap.protocolversion
  leng += msix_Init(&msrv->msix, 1);        // dmap.supportsindex
  leng += msex_Init(&msrv->msex, 1);        // dmap.supportsextensions
  leng += msup_Init(&msrv->msup, 1);        // dmap.supportsupdate
  leng += msal_Init(&msrv->msal, 1);        // dmap.supportsautologout
  leng += mstm_Init(&msrv->mstm, 1800);     // dmap.timeoutinterval
  leng += mslr_Init(&msrv->mslr, 0);        // dmap.loginrequired
  leng += msqy_Init(&msrv->msqy, 1);        // dmap.supportsquery
  leng += minm_Init(&msrv->minm, "iCast Music");  // dmap.itemname
//leng += minm_Init(&msrv->minm, "iBook Music");  // dmap.itemname
  leng += msrs_Init(&msrv->msrs, 1);        // dmap.supportsresolve
  leng += msbr_Init(&msrv->msbr, 1);        // dmap.supportsbrowse
  leng += mspi_Init(&msrv->mspi, 1);        // dmap.supportspersistentids
  leng += msdc_Init(&msrv->msdc, 1);        // dmap.databasescount
  
  msrv->base.name = 0x6d737276;
  msrv->base.leng = leng;
}

buff_t *
msrv_Serialize (daap_msrv * msrv)
{
  buff_t * buff;
  daap_int name = htonl(msrv->base.name);
  daap_int leng = 0;
  
  assert(msrv != NULL);

  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);

  buff_Append(buff, &name, 4);         // msrv name
  buff_Append(buff, &leng, 4);         // placeholder for length
  
  daap_Serialize((daap_base *)&msrv->mstt, buff);
  daap_Serialize((daap_base *)&msrv->mpro, buff);
  daap_Serialize((daap_base *)&msrv->apro, buff);
  daap_Serialize((daap_base *)&msrv->msix, buff);
  daap_Serialize((daap_base *)&msrv->msex, buff);
  daap_Serialize((daap_base *)&msrv->msup, buff);
  daap_Serialize((daap_base *)&msrv->msal, buff);
  daap_Serialize((daap_base *)&msrv->mstm, buff);
  daap_Serialize((daap_base *)&msrv->mslr, buff);
  daap_Serialize((daap_base *)&msrv->msqy, buff);
  daap_Serialize((daap_base *)&msrv->minm, buff);
  daap_Serialize((daap_base *)&msrv->msrs, buff);
  daap_Serialize((daap_base *)&msrv->msbr, buff);
  daap_Serialize((daap_base *)&msrv->mspi, buff);
  daap_Serialize((daap_base *)&msrv->msdc, buff);
  
  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
msrv_Destroy (daap_msrv * msrv)
{
  assert(msrv != NULL);
  
  if (msrv->minm.valu != NULL)
    free(msrv->minm.valu);
  
  free(msrv);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 02 [mccr] /content-codes
#endif

daap_mccr *
mccr_Create ()
{
  daap_mccr * mccr;
  
  mccr = (daap_mccr *)malloc(sizeof(daap_mccr));
  assert(mccr != NULL);
  
  return mccr;
}

void
mccr_Init (daap_mccr * mccr)
{
  daap_int leng = 0;
  int i = 0;
  daap_mdcl * mdcl;
  daap_mcna * mcna;
  daap_mcnm * mcnm;
  daap_mcty * mcty;
  daap_int name = 0;
  
  assert(mccr != NULL);

  bzero(mccr, sizeof(daap_mccr));
  
  leng += mstt_Init(&mccr->mstt, 200);      // dmap.status
  
  mccr->mdcl.len  = 0;
  mccr->mdcl.max  = 0;
  mccr->mdcl.list = NULL;

  for (i = 0; i < 85; i++) {
    
    mdcl = (daap_mdcl *)malloc(sizeof(daap_mdcl));
    assert(mdcl != NULL);
    bzero(mdcl, sizeof(daap_mdcl));
    
    mcna = (daap_mcna *)malloc(sizeof(daap_mcna));
    assert(mcna != NULL);
    bzero(mcna, sizeof(daap_mcna));
    
    mcnm = (daap_mcnm *)malloc(sizeof(daap_mcnm));
    assert(mcnm != NULL);
    bzero(mcnm, sizeof(daap_mcnm));
    
    mcty = (daap_mcty *)malloc(sizeof(daap_mcty));
    assert(mcty != NULL);
    bzero(mcty, sizeof(daap_mcty));
    
    memcpy(&name, ((daap_type)daap_type_list[i]).name, 4);
    
    mdcl_Init(mdcl);
    mcna_Init(mcna,           ((daap_type *)&daap_type_list[i])->path);         // dmap.status
    mcnm_Init(mcnm, name);          // mstt
    mcty_Init(mcty,           ((daap_type *)&daap_type_list[i])->type);         // 0x0005
    
    daap_list_Append((__daap_list *)mdcl, (daap_base *)mcna);
    daap_list_Append((__daap_list *)mdcl, (daap_base *)mcnm);
    daap_list_Append((__daap_list *)mdcl, (daap_base *)mcty);
    
    mccr_AddContentCode(mccr, mdcl);
  }
  
  mccr->base.name = 0x6d636372;
  mccr->base.leng = leng;
}

void
mccr_AddContentCode (daap_mccr * mccr, daap_mdcl * mdcl)
{
  assert(mccr != NULL);
  assert(mdcl != NULL);

  if (mccr->mdcl.list == NULL) {
    mccr->mdcl.max  = 5;
    mccr->mdcl.list = (void **)malloc(sizeof(void *) * mccr->mdcl.max);
  }
  else if (mccr->mdcl.max <= mccr->mdcl.len) {
    mccr->mdcl.max *= 1.5;
    mccr->mdcl.list = (void **)realloc(mccr->mdcl.list, sizeof(void *) * mccr->mdcl.max);
  }

  assert(mccr->mdcl.list != NULL);
  
  mccr->mdcl.list[mccr->mdcl.len++] = mdcl;
}

buff_t *
mccr_Serialize (daap_mccr * mccr)
{
  buff_t * buff;
  int i = 0;
  daap_int name;
  daap_int leng;
  
  assert(mccr != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(mccr->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // mccr name
  buff_Append(buff, &leng, 4);         // placeholder for length

  daap_Serialize((daap_base *)&mccr->mstt, buff);

  for (i = 0; i < mccr->mdcl.len; i++) {
    /**
    int x = 0;
    int y = 0;
     
    fprintf(stderr, "\n\n\nmccr_Serialize.. buff->len = %d\n", buff->len);
    for (x = 0; x < buff->len;) {
      for (y = 0; y < 8 && x < buff->len; y++, x++)
        fprintf(stderr, "%02hhx ", (char)*(buff->buf + x));
      fprintf(stderr, "   ");
      for (y = 0; y < 8 && x < buff->len; y++, x++)
        fprintf(stderr, "%02hhx ", (char)*(buff->buf + x));
      fprintf(stderr, "\n");
    }
    fprintf(stderr, "\n\n\n");
    **/
    
    daap_Serialize( (daap_base *)(daap_mdcl *)mccr->mdcl.list[i], buff);
  }

  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
mccr_Destroy (daap_mccr * mccr)
{
  int i = 0;
  
  assert(mccr != NULL);

  for (i = 0; i < mccr->mdcl.len; i++)
//  daap_Destroy((daap_base *)&mccr->mdcl.list[i]);
    daap_Destroy( (daap_base *)(daap_mdcl *)mccr->mdcl.list[i]);
  
//daap_Destroy((daap_base *)&mccr->mdcl);
  
  free(mccr);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 03 [mlog] /login
#endif

daap_mlog *
mlog_Create ()
{
  daap_mlog * mlog;
  
  mlog = (daap_mlog *)malloc(sizeof(daap_mlog));
  assert(mlog != NULL);
  
  return mlog;
}

void
mlog_Init (daap_mlog * mlog)
{
  daap_int leng = 0;
  
  assert(mlog != NULL);
  
  bzero(mlog, sizeof(daap_mlog));
  
  leng += mstt_Init(&mlog->mstt, 200);      // dmap.status
  leng += mlid_Init(&mlog->mlid, 0x1234);   // dmap.sessionid
  
  mlog->base.name = 0x6d6c6f67;
  mlog->base.leng = leng;
}

buff_t *
mlog_Serialize (daap_mlog * mlog)
{
  buff_t * buff;
  daap_int name;
  daap_int leng;
  
  assert(mlog != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(mlog->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // msrv name
  buff_Append(buff, &leng, 4);         // placeholder for length
  
  daap_Serialize((daap_base *)&mlog->mstt, buff);
  daap_Serialize((daap_base *)&mlog->mlid, buff);
  
  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
mlog_Destroy (daap_mlog * mlog)
{
  assert(mlog != NULL);
  
  free(mlog);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 04 [mupd] /update
#endif

daap_mupd *
mupd_Create ()
{
  daap_mupd * mupd;
  
  mupd = (daap_mupd *)malloc(sizeof(daap_mupd));
  assert(mupd != NULL);
  
  return mupd;
}

void
mupd_Init (daap_mupd * mupd)
{
  daap_int leng = 0;
  
  assert(mupd != NULL);
  
  bzero(mupd, sizeof(daap_mupd));
  
  leng += mstt_Init(&mupd->mstt, 200);      // dmap.status
  leng += musr_Init(&mupd->musr, 3);        // dmap.sessionid
  
  mupd->base.name = 0x6d757064;
  mupd->base.leng = leng;
}

buff_t *
mupd_Serialize (daap_mupd * mupd)
{
  buff_t * buff;
  daap_int name;
  daap_int leng;
  
  assert(mupd != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(mupd->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // msrv name
  buff_Append(buff, &leng, 4);         // placeholder for length
  
  daap_Serialize((daap_base *)&mupd->mstt, buff);
  daap_Serialize((daap_base *)&mupd->musr, buff);
  
  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
mupd_Destroy (daap_mupd * mupd)
{
  assert(mupd != NULL);
  
  free(mupd);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 05 [avdb] /databases
#endif

daap_avdb *
avdb_Create ()
{
  daap_avdb * avdb;
  
  avdb = (daap_avdb *)malloc(sizeof(daap_avdb));
  assert(avdb != NULL);
  
  return avdb;
}

void
avdb_Init (daap_avdb * avdb)
{
  daap_int leng = 0;
  daap_mlit * mlit;
  daap_miid * miid;
  daap_mper * mper;
  daap_minm * minm;
  daap_mimc * mimc;
  daap_mctc * mctc;
  
  assert(avdb != NULL);

  bzero(avdb, sizeof(daap_avdb));
  
  leng += mstt_Init(&avdb->mstt, 200);      // dmap.status
  leng += muty_Init(&avdb->muty, 0);        // dmap.updatetype
  leng += mtco_Init(&avdb->mtco, 1);        // dmap.specifiedtotalcount
  leng += mrco_Init(&avdb->mrco, 1);        // dmap.returnedcount
  leng += mlcl_Init(&avdb->mlcl);           // dmap.listing

  mlit = (daap_mlit *)malloc(sizeof(daap_mlit));
  assert(mlit != NULL);
  mlit_Init(mlit);
  
  {
    // database id
    miid = (daap_miid *)malloc(sizeof(daap_miid));
    assert(miid != NULL);
    bzero(miid, sizeof(daap_miid));
    miid_Init(miid, 0x20);
    
    // database persistent id
    mper = (daap_mper *)malloc(sizeof(daap_mper));
    assert(mper != NULL);
    bzero(mper, sizeof(daap_mper));
    mper_Init(mper, 0x73a30788);
    
    // database name
    minm = (daap_minm *)malloc(sizeof(daap_minm));
    assert(minm != NULL);
    bzero(minm, sizeof(daap_minm));
    minm_Init(minm, "iCast Music");
    
    // number of songs in this database
    mimc = (daap_mimc *)malloc(sizeof(daap_mimc));
    assert(mimc != NULL);
    bzero(mimc, sizeof(daap_mimc));
    mimc_Init(mimc, kDefaultDatabase.len);
    
    // number of playlists in this database
    mctc = (daap_mctc *)malloc(sizeof(daap_mctc));
    assert(mctc != NULL);
    bzero(mctc, sizeof(daap_mctc));
    mctc_Init(mctc, 1);
    
    daap_list_Append(mlit, (daap_base *)miid);
    daap_list_Append(mlit, (daap_base *)mper);
    daap_list_Append(mlit, (daap_base *)minm);
    daap_list_Append(mlit, (daap_base *)mimc);
    daap_list_Append(mlit, (daap_base *)mctc);
  }

  daap_list_Append(&avdb->mlcl, (daap_base *)mlit);
  
  avdb->base.name = 0x61766462;
  avdb->base.leng = leng;
}

void
avdb_AddRecord (daap_avdb * avdb, daap_mdcl * mdcl)
{
  assert(avdb != NULL);
  assert(mdcl != NULL);

  (void)avdb;
  (void)mdcl;
}

buff_t *
avdb_Serialize (daap_avdb * avdb)
{
  buff_t * buff;
  daap_int name;
  daap_int leng;
  
  assert(avdb != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(avdb->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // avdb name
  buff_Append(buff, &leng, 4);         // placeholder for length

  daap_Serialize((daap_base *)&avdb->mstt, buff);
  daap_Serialize((daap_base *)&avdb->muty, buff);
  daap_Serialize((daap_base *)&avdb->mtco, buff);
  daap_Serialize((daap_base *)&avdb->mrco, buff);
  daap_Serialize((daap_base *)&avdb->mlcl, buff);

  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
avdb_Destroy (daap_avdb * avdb)
{
  int i = 0;
  
  assert(avdb != NULL);

  for (i = 0; i < avdb->mlcl.valu.len; i++)
    daap_Destroy( (daap_base *)(daap_mlit *)avdb->mlcl.valu.list[i]);
  
  free(avdb);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 06 [adbs] /databases/<dbid>/items
#endif

daap_adbs *
adbs_Create ()
{
  daap_adbs * adbs;
  
  adbs = (daap_adbs *)malloc(sizeof(daap_adbs));
  assert(adbs != NULL);
  
  return adbs;
}

void
adbs_Init (daap_adbs * adbs)
{
  daap_int leng = 0;
  
  assert(adbs != NULL);
  
  bzero(adbs, sizeof(daap_adbs));
  
  leng += mstt_Init(&adbs->mstt, 200);      // dmap.status
  leng += muty_Init(&adbs->muty, 0);        // dmap.updatetype
  leng += mtco_Init(&adbs->mtco, kDefaultDatabase.len);        // dmap.specifiedtotalcount
  leng += mrco_Init(&adbs->mrco, kDefaultDatabase.len);        // dmap.returnedcount
  leng += mlcl_Init(&adbs->mlcl);           // dmap.listing
  
  // adbs_LoadDir(adbs, "./files");
  
  adbs->base.name = 0x61646273;
  adbs->base.leng = leng;
}

void 
adbs_AddSong (adbs, song)
  daap_adbs * adbs;
  daap_song * song;
{
  daap_mlit * mlit;
  daap_mikd * mikd;
  daap_miid * miid;
  daap_minm * minm;
  daap_mper * mper;
  daap_asal * asal;
  daap_asar * asar;
  daap_asbr * asbr;
  daap_asbt * asbt;
  daap_ascm * ascm;
  daap_asco * asco;
//daap_ascp * ascp;
  daap_asdc * asdc;
  daap_asdn * asdn;
  daap_asdb * asdb;
  daap_aseq * aseq;
  daap_asfm * asfm;
  daap_asgn * asgn;
  daap_asdt * asdt;
//daap_asrv * asrv;
  daap_assr * assr;
  daap_assz * assz;
  daap_asdk * asdk;
  daap_aeNV * aeNV;
  daap_aeSP * aeSP;
  
  assert(adbs != NULL);
  assert(song != NULL);

  mlit = (daap_mdcl *)malloc(sizeof(daap_mlit));
  mlit_Init(mlit);

  mikd = (daap_mikd *)malloc(sizeof(daap_mikd));
  mikd_Init(mikd, song->mikd.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)mikd);
  
  miid = (daap_miid *)malloc(sizeof(daap_miid));
  miid_Init(miid, song->miid.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)miid);
  
  minm = (daap_minm *)malloc(sizeof(daap_minm));
  minm_Init(minm, song->minm.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)minm);

  mper = (daap_mper *)malloc(sizeof(daap_mper));
  mper_Init(mper, song->mper.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)mper);

  asal = (daap_asal *)malloc(sizeof(daap_asal));
  asal_Init(asal, song->asal.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asal);

  asar = (daap_asar *)malloc(sizeof(daap_asar));
  asar_Init(asar, song->asar.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asar);
  
  asbr = (daap_asbr *)malloc(sizeof(daap_asbr));
  asbr_Init(asbr, song->asbr.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asbr);

  asbt = (daap_asbt *)malloc(sizeof(daap_asbt));
  asbt_Init(asbt, song->asbt.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asbt);

  ascm = (daap_ascm *)malloc(sizeof(daap_ascm));
  ascm_Init(ascm, song->ascm.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)ascm);

  asco = (daap_asco *)malloc(sizeof(daap_asco));
  asco_Init(asco, song->asco.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asco);

  // ascp = (daap_ascp *)malloc(sizeof(daap_ascp));
  // ascp_Init(ascp, song->ascp.valu);
  // daap_list_Append((__daap_list *)mlit, (daap_base *)ascp);
  
  asdc = (daap_asdc *)malloc(sizeof(daap_asdc));
  asdc_Init(asdc, song->asdc.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asdc);
  
  asdn = (daap_asdn *)malloc(sizeof(daap_asdn));
  asdn_Init(asdn, song->asdn.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asdn);
  
  asdb = (daap_asdb *)malloc(sizeof(daap_asdb));
  asdb_Init(asdb, song->asdb.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asdb);
  
  aseq = (daap_aseq *)malloc(sizeof(daap_aseq));
  aseq_Init(aseq, song->aseq.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)aseq);
  
  asfm = (daap_asfm *)malloc(sizeof(daap_asfm));
  asfm_Init(asfm, song->asfm.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asfm);

  asgn = (daap_asgn *)malloc(sizeof(daap_asgn));
  asgn_Init(asgn, song->asgn.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asgn);
  
  asdt = (daap_asdt *)malloc(sizeof(daap_asdt));
  asdt_Init(asdt, song->asdt.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asdt);
  
  // asrv = (daap_asrv *)malloc(sizeof(daap_asrv));
  // asrv_Init(asrv, song->asrv.valu);
  // daap_list_Append((__daap_list *)mlit, (daap_base *)asrv);
  
  assr = (daap_assr *)malloc(sizeof(daap_assr));
  assr_Init(assr, song->assr.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)assr);
  
  assz = (daap_assz *)malloc(sizeof(daap_assz));
  assz_Init(assz, song->assz.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)assz);
  
  asdk = (daap_asdk *)malloc(sizeof(daap_asdk));
  asdk_Init(asdk, song->asdk.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)asdk);

  aeNV = (daap_aeNV *)malloc(sizeof(daap_aeNV));
  aeNV_Init(aeNV, song->aeNV.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)aeNV);
  
  aeSP = (daap_aeSP *)malloc(sizeof(daap_aeSP));
  aeSP_Init(aeSP, song->aeSP.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)aeSP);

  
  fprintf(stderr, "adbs_AddSong.. adding song, '%s', '%d', '%d'\n", song->minm.valu, song->miid.valu, (int)song->mper.valu);
  
  adbs_AddRecord(adbs, mlit);
}

void
adbs_AddRecord (daap_adbs * adbs, daap_mlit * mlit)
{
  assert(adbs != NULL);
  assert(mlit != NULL);
  
  daap_list_Append((__daap_list *)&adbs->mlcl, (daap_base *)mlit);
}

buff_t *
adbs_Serialize (daap_adbs * adbs)
{
  buff_t * buff;
  daap_int name;
  daap_int leng;
  
  assert(adbs != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(adbs->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // adbs name
  buff_Append(buff, &leng, 4);         // placeholder for length
  
  daap_Serialize((daap_base *)&adbs->mstt, buff);
  daap_Serialize((daap_base *)&adbs->muty, buff);
  daap_Serialize((daap_base *)&adbs->mtco, buff);
  daap_Serialize((daap_base *)&adbs->mrco, buff);
  daap_Serialize((daap_base *)&adbs->mlcl, buff);
  
  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
adbs_Destroy (daap_adbs * adbs)
{
  int i = 0;
  
  assert(adbs != NULL);
  
  for (i = 0; i < adbs->mlcl.valu.len; i++)
    daap_Destroy( (daap_base *)(daap_mlit *)adbs->mlcl.valu.list[i]);
  
  free(adbs);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 07 [aply] /databases/<dbid>/containers
#endif

daap_aply *
aply_Create ()
{
  daap_aply * aply;
  
  aply = (daap_aply *)malloc(sizeof(daap_aply));
  assert(aply != NULL);
  
  return aply;
}

void
aply_Init (daap_aply * aply)
{
  daap_int leng = 0;
  
  assert(aply != NULL);
  
  bzero(aply, sizeof(daap_aply));
  
  leng += mstt_Init(&aply->mstt, 200);      // dmap.status
  leng += muty_Init(&aply->muty, 0);        // dmap.updatetype
  leng += mtco_Init(&aply->mtco, 1);        // dmap.specifiedtotalcount
  leng += mrco_Init(&aply->mrco, 1);        // dmap.returnedcount
  leng += mlcl_Init(&aply->mlcl);           // dmap.listing

  aply->base.name = 0x61706c79;
  aply->base.leng = leng;
  
  {
    daap_playlist playlist;
    bzero(&playlist, sizeof(daap_playlist));

    miid_Init(&playlist.miid, 0x01);
    mper_Init(&playlist.mper, 0x01);
    minm_Init(&playlist.minm, "Playlist");
    mimc_Init(&playlist.mimc, kDefaultDatabase.len);
    aeSP_Init(&playlist.aeSP, 0x00);
    
    aply_AddPlaylist(aply, &playlist);
  }
}

void 
aply_AddPlaylist (aply, playlist)
  daap_aply * aply;
daap_playlist * playlist;
{
  daap_mlit * mlit;
  daap_miid * miid;
  daap_mper * mper;
  daap_minm * minm;
  daap_mimc * mimc;
  daap_aeSP * aeSP;
  
  assert(aply != NULL);
  assert(playlist != NULL);
  
  mlit = (daap_mdcl *)malloc(sizeof(daap_mlit));
  mlit_Init(mlit);
  
  miid = (daap_miid *)malloc(sizeof(daap_miid));
  assert(miid != NULL);
  bzero(miid, sizeof(daap_miid));
  miid_Init(miid, playlist->miid.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)miid);
  
  mper = (daap_mper *)malloc(sizeof(daap_mper));
  assert(mper != NULL);
  bzero(mper, sizeof(daap_mper));
  mper_Init(mper, playlist->mper.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)mper);
  
  minm = (daap_minm *)malloc(sizeof(daap_minm));
  assert(minm != NULL);
  bzero(minm, sizeof(daap_minm));
  minm_Init(minm, playlist->minm.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)minm);

  mimc = (daap_mimc *)malloc(sizeof(daap_mimc));
  assert(mimc != NULL);
  bzero(mimc, sizeof(daap_mimc));
  mimc_Init(mimc, playlist->mimc.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)mimc);

  aeSP = (daap_aeSP *)malloc(sizeof(daap_aeSP));
  assert(aeSP != NULL);
  bzero(aeSP, sizeof(daap_aeSP));
  aeSP_Init(aeSP, playlist->aeSP.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)aeSP);
  
  aply_AddRecord(aply, mlit);
}

void
aply_AddRecord (daap_aply * aply, daap_mlit * mlit)
{
  assert(aply != NULL);
  assert(mlit != NULL);
  
  daap_list_Append((__daap_list *)&aply->mlcl, (daap_base *)mlit);
}

buff_t *
aply_Serialize (daap_aply * aply)
{
  buff_t * buff;
  daap_int name;
  daap_int leng;
  
  assert(aply != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(aply->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // aply name
  buff_Append(buff, &leng, 4);         // placeholder for length
  
  daap_Serialize((daap_base *)&aply->mstt, buff);
  daap_Serialize((daap_base *)&aply->muty, buff);
  daap_Serialize((daap_base *)&aply->mtco, buff);
  daap_Serialize((daap_base *)&aply->mrco, buff);
  daap_Serialize((daap_base *)&aply->mlcl, buff);
  
  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
aply_Destroy (daap_aply * aply)
{
  int i = 0;
  
  assert(aply != NULL);
  
  for (i = 0; i < aply->mlcl.valu.len; i++)
    daap_Destroy( (daap_base *)(daap_mlit *)aply->mlcl.valu.list[i]);
  
  free(aply);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** step 08 [apso] /databases/<dbid>/containers/<plid>/items
#endif

daap_apso *
apso_Create ()
{
  daap_apso * apso;
  
  apso = (daap_apso *)malloc(sizeof(daap_apso));
  assert(apso != NULL);
  
  return apso;
}

void
apso_Init (daap_apso * apso)
{
  daap_int leng = 0;
  
  assert(apso != NULL);
  
  bzero(apso, sizeof(daap_apso));
  
  leng += mstt_Init(&apso->mstt, 200);      // dmap.status
  leng += muty_Init(&apso->muty, 0);        // dmap.updatetype
  leng += mtco_Init(&apso->mtco, 0);        // dmap.specifiedtotalcount
  leng += mrco_Init(&apso->mrco, 0);        // dmap.returnedcount
  leng += mlcl_Init(&apso->mlcl);           // dmap.listing
  
  apso->base.name = 0x6170736f;
  apso->base.leng = leng;
}

void 
apso_AddSong (apso, song)
  daap_apso * apso;
  daap_song * song;
{
  daap_mlit * mlit;
  daap_mikd * mikd;
  daap_miid * miid;
  daap_mcti * mcti;
  
  assert(apso != NULL);
  assert(song != NULL);
  
  mlit = (daap_mdcl *)malloc(sizeof(daap_mlit));
  mlit_Init(mlit);

  mikd = (daap_mikd *)malloc(sizeof(daap_mikd));
  mikd_Init(mikd, song->mikd.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)mikd);
  
  miid = (daap_miid *)malloc(sizeof(daap_miid));
  miid_Init(miid, song->miid.valu);
  daap_list_Append((__daap_list *)mlit, (daap_base *)miid);

  mcti = (daap_mcti *)malloc(sizeof(daap_mcti));
  mcti_Init(mcti, 0x01);
  daap_list_Append((__daap_list *)mlit, (daap_base *)mcti);

  apso_AddRecord(apso, mlit);
}

void
apso_AddRecord (daap_apso * apso, daap_mlit * mlit)
{
  assert(apso != NULL);
  assert(mlit != NULL);
  
  daap_list_Append((__daap_list *)&apso->mlcl, (daap_base *)mlit);
}

buff_t *
apso_Serialize (daap_apso * apso)
{
  buff_t * buff;
  daap_int name;
  daap_int leng;
  
  assert(apso != NULL);
  
  buff = buff_Create();
  buff_Init(buff, BUFF_UNLIMITED);
  
  name = htonl(apso->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);         // apso name
  buff_Append(buff, &leng, 4);         // placeholder for length
  
  daap_Serialize((daap_base *)&apso->mstt, buff);
  daap_Serialize((daap_base *)&apso->muty, buff);
  daap_Serialize((daap_base *)&apso->mtco, buff);
  daap_Serialize((daap_base *)&apso->mrco, buff);
  daap_Serialize((daap_base *)&apso->mlcl, buff);
  
  leng = htonl(buff->leng - 8);
  memcpy(buff->data + 4, &leng, 4);
  
  return buff;
}

void
apso_Destroy (daap_apso * apso)
{
  int i = 0;
  
  assert(apso != NULL);
  
  for (i = 0; i < apso->mlcl.valu.len; i++)
    daap_Destroy( (daap_base *)(daap_mlit *)apso->mlcl.valu.list[i]);
  
  free(apso);
}







#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Song & Database Methods
#endif

daap_song *
daap_song_Create ()
{
  daap_song * song;
  
  song = (daap_song *)malloc(sizeof(daap_song));
  assert(song != NULL);
  
  return song;
}

void
daap_song_Init (song, name)
  daap_song * song;
  daap_int    name;
{
  assert(song != NULL);

  bzero(song, sizeof(daap_song));

  song->file_name = NULL;
  song->file_path = NULL;

  song->base.name = name;
  song->base.leng = 0;
  song->base.type = DAAP_TYPE_SONG;
  
  /**
  bzero(&song->miid, sizeof(daap_miid));
  bzero(&song->minm, sizeof(daap_minm));
  bzero(&song->mikd, sizeof(daap_mikd));
  bzero(&song->mper, sizeof(daap_mper));
  bzero(&song->asal, sizeof(daap_asal));
  bzero(&song->agrp, sizeof(daap_agrp));
  bzero(&song->asar, sizeof(daap_asar));
  bzero(&song->asbr, sizeof(daap_asbr));
  bzero(&song->asbt, sizeof(daap_asbt));
  bzero(&song->ascm, sizeof(daap_ascm));
  bzero(&song->asco, sizeof(daap_asco));
  bzero(&song->ascp, sizeof(daap_ascp));
  bzero(&song->asda, sizeof(daap_asda));
  bzero(&song->asdm, sizeof(daap_asdm));
  bzero(&song->asdc, sizeof(daap_asdc));
  bzero(&song->asdn, sizeof(daap_asdn));
  bzero(&song->asdb, sizeof(daap_asdb));
  bzero(&song->aseq, sizeof(daap_aseq));
  bzero(&song->asfm, sizeof(daap_asfm));
  bzero(&song->asgn, sizeof(daap_asgn));
  bzero(&song->asdt, sizeof(daap_asdt));
  bzero(&song->asrv, sizeof(daap_asrv));
  bzero(&song->assr, sizeof(daap_assr));
  bzero(&song->assz, sizeof(daap_assz));
  bzero(&song->asst, sizeof(daap_asst));
  bzero(&song->assp, sizeof(daap_assp));
  bzero(&song->astm, sizeof(daap_astm));
  bzero(&song->astc, sizeof(daap_astc));
  bzero(&song->astn, sizeof(daap_astn));
  bzero(&song->asur, sizeof(daap_asur));
  bzero(&song->asyr, sizeof(daap_asyr));
  bzero(&song->asdk, sizeof(daap_asdk));
  bzero(&song->asul, sizeof(daap_asul));
  bzero(&song->aenv, sizeof(daap_aenv));
  **/
}

daap_int
daap_song_Serialize (song, buff)
  daap_song * song;
  buff_t    * buff;
{
  daap_int name;
  daap_int leng;
  int pos;
  
  assert(song != NULL);
  assert(buff != NULL);
  
  song->base.leng = 0;
  
  name = htonl(song->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  
  pos = buff->leng;
  
  song->base.leng += daap_Serialize((daap_base *)&song->mikd, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->miid, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->minm, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->mper, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asal, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->agrp, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asar, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asbr, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asbt, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->ascm, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asco, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->ascp, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asda, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asdm, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asdc, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asdn, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asdb, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->aseq, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asfm, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asgn, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asdt, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asrv, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->assr, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->assz, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asst, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->assp, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->astm, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->astc, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->astn, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asur, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asyr, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asdk, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->asul, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->aeNV, buff);
  song->base.leng += daap_Serialize((daap_base *)&song->aeSP, buff);
  
  leng = htonl(song->base.leng);
  memcpy(buff->data + (pos - 4), &leng, 4);

  return 8 + song->base.leng;
}

void
daap_song_Destroy (song)
  daap_song * song;
{
  assert(song != NULL);
  
  if (song->file_name != NULL)
    free(song->file_name);
  
  if (song->file_path != NULL)
    free(song->file_path);
  
  song->file_name = NULL;
  song->file_path = NULL;

  // TODO: go through to each one of the members of this structure
  // which is of type DAAP_TYPE_STRING, and free() the string if
  // any room for one has been malloc()'ed.
  
  free(song);
}

daap_database *
daap_database_Create ()
{
  daap_database * database;
  
  database = (daap_database *)malloc(sizeof(daap_database));
  assert(database != NULL);
  
  return database;
}

void
daap_database_Init (database)
  daap_database * database;
{
  int status = 0;
  
  assert(database != NULL);
  
  status = pthread_mutex_init(&database->mutex, NULL);
  assert(status == 0);

  database->len   = 0;
  database->max   = 5;
  database->songs = NULL;
  database->files = NULL;
  
  database->files = hash_Create();
  hash_Init(database->files);
  
  database->songs = (daap_song **)malloc(sizeof(daap_song *) * database->max);
  assert(database->songs != NULL);
}

//
// import, verify and update an existing database which
// has been Export()'ed previously. the process of verifying
// involves checking to see that all of the songs in the database
// still exist on disk, with the same file size, and then it
// checks to see if there are new files on disk, which are not
// in the database, and adds them, if appropriate.
//
daap_int
daap_database_Import (database)
  daap_database * database;
{
  FILE * file;
  int count = 0;
  daap_song * song;
  daap_int sname = 0;               // song name
  daap_int sleng = 0;               // song length
  unsigned char data[1000];
  daap_int fpos = 0;
  daap_int songid = 0;
  
  assert(database != NULL);
  
  if (NULL == (file = fopen("database.icast", "r"))) {
    fprintf(stderr, "daap_database_Import.. could not locate database file\n");
    return -1;
  }
  
  while (1)
  {
    if (4 != fread(&sname, 1, 4, file))
      break;
    
    if (4 != fread(&sleng, 1, 4, file)) {
      fprintf(stderr, "daap_database_Import.. failed to read what should be the song length [1]\n");
      break;
    }

    count++;
    
    // create and initialize a now daap_song
    song = daap_song_Create();
    daap_song_Init(song, 0x6d6c6974);
    songid = daap_database_AppendSong(database, song);
    miid_Init(&song->miid, songid + 1);

    // fix the endianess of these values
    sname = ntohl(sname);
    sleng = ntohl(sleng);

    // read in the data for the entire long
    assert(sleng <= 1000);
    fread(&data, 1, sleng, file);
    
    for (fpos = 0; fpos < sleng;) {
      daap_int name = 0;
      daap_int leng = 0;
      
      memcpy(&name, data + fpos + 0, 4);
      memcpy(&leng, data + fpos + 4, 4);
      
      name = ntohl(name);
      leng = ntohl(leng);

      switch (name) {
        case 0x736f6e67:                      // song
          fprintf(stderr, "daap_database_Import.. %d: formatting error at %d\n", count, fpos);
          break;
        case 0x70617468:                      // path
          song->file_path = (char *)malloc(leng + 1);
          bzero(song->file_path, leng + 1);
          song->file_path[leng] = 0;
          memcpy(song->file_path, data + fpos + 8, leng);
          hash_Put(database->files, song->file_path, leng, (char *)&songid, 4);
          fprintf(stderr, "daap_database_Import.. %04d: file_path = %s\n", count, song->file_path);
          break;
        case 0x6e616d65:                      // name
          break;
        case 0x6d696b64:                      // mikd
          daap_Import((daap_base *)&song->mikd, DAAP_TYPE_BYTE,   data + fpos);
          break;
        case 0x6d696964:                      // miid
          break;
        case 0x6d696e6d:                      // minm
          daap_Import((daap_base *)&song->minm, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x6d706572:                      // mper
          daap_Import((daap_base *)&song->mper, DAAP_TYPE_LONG,   data + fpos);
          break;
        case 0x6173616c:                      // asal
          daap_Import((daap_base *)&song->asal, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x61736172:                      // asar
          daap_Import((daap_base *)&song->asar, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x61736272:                      // asbr
          daap_Import((daap_base *)&song->asbr, DAAP_TYPE_SHORT,  data + fpos);
          break;
        case 0x61736274:                      // asbt
          daap_Import((daap_base *)&song->asbt, DAAP_TYPE_SHORT,  data + fpos);
          break;
        case 0x6173636d:                      // ascm
          daap_Import((daap_base *)&song->ascm, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x6173636f:                      // asco
          daap_Import((daap_base *)&song->asco, DAAP_TYPE_BYTE,   data + fpos);
          break;
        case 0x61736370:                      // ascp
          break;
//      case 0x:                              // asda
//        break;
        case 0x6173646d:                      // asdm
          daap_Import((daap_base *)&song->asdm, DAAP_TYPE_INT,    data + fpos);
          break;
        case 0x61736463:                      // asdc
          daap_Import((daap_base *)&song->asdc, DAAP_TYPE_SHORT,  data + fpos);
          break;
        case 0x6173646e:                      // asdn
          daap_Import((daap_base *)&song->asdn, DAAP_TYPE_SHORT,  data + fpos);
          break;
        case 0x61736462:                      // asdb
          daap_Import((daap_base *)&song->asdb, DAAP_TYPE_BYTE,   data + fpos);
          break;
        case 0x61736571:                      // aseq
          daap_Import((daap_base *)&song->aseq, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x6173666d:                      // asfm
          daap_Import((daap_base *)&song->asfm, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x6173676e:                      // asgn
          daap_Import((daap_base *)&song->asgn, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x61736474:                      // asdt
          daap_Import((daap_base *)&song->asdt, DAAP_TYPE_STRING, data + fpos);
          break;
        case 0x61737276:                      // asrv
          break;
        case 0x61737372:                      // assr
          daap_Import((daap_base *)&song->assr, DAAP_TYPE_INT,    data + fpos);
          break;
        case 0x6173737a:                      // assz
          daap_Import((daap_base *)&song->assz, DAAP_TYPE_INT,    data + fpos);
          break;
//      case 0x:                              // asst
//        break;
//      case 0x.                              // assp
//        break;
        case 0x6173746d:                      // astm
          daap_Import((daap_base *)&song->astm, DAAP_TYPE_INT,    data + fpos);
          break;
        case 0x61737463:                      // astc
          daap_Import((daap_base *)&song->astc, DAAP_TYPE_SHORT,  data + fpos);
          break;
        case 0x6173746e:                      // astn
          daap_Import((daap_base *)&song->astn, DAAP_TYPE_SHORT,  data + fpos);
          break;
//      case 0x:                              // asur
//        break;
        case 0x61737972:                      // asyr
          daap_Import((daap_base *)&song->asyr, DAAP_TYPE_SHORT,  data + fpos);
          break;
        case 0x6173646b:                      // asdk
          daap_Import((daap_base *)&song->asdk, DAAP_TYPE_BYTE,   data + fpos);
          break;
//      case 0x:                              // asul
//        break;
        case 0x61654e56:                      // aeNV
          daap_Import((daap_base *)&song->aeNV, DAAP_TYPE_INT,    data + fpos);
          break;
        case 0x61655350:                      // aeSP
          daap_Import((daap_base *)&song->aeSP, DAAP_TYPE_BYTE,   data + fpos);
          break;
        default:
          fprintf(stderr, "daap_database_Import.. %d: unsupported type, %s\n", count, (char *)&name);
          break;
      }
      
      fpos += 4 + 4 + leng;
    }
  }

  /**
  // if we added or deleted any entries, update the persistent database accordingly
  if (0 < daap_database_DelOldEntries(database) || 0 < daap_database_AddNewEntries(database))
    daap_database_Export(database);
  **/
  
  fclose(file);
  
  return count;
}

daap_int
daap_database_Update (database, path)
  daap_database * database;
  const char * path;
{
  int dcount = 0;               // entry deletion count
  int acount = 0;               // entry addition count
  
  assert(database != NULL);
  assert(path     != NULL);
  
  dcount = daap_database_DelOldEntries(database);
  acount = daap_database_LoadDir(database, path);

  fprintf(stderr, "daap_database_Update.. %s, dcount = %d\n", path, dcount);
  fprintf(stderr, "daap_database_Update.. %s, acount = %d\n", path, acount);
  
  if (dcount > 0 || acount > 0)
    daap_database_Export(database);

  return dcount + acount;
}

daap_int
daap_database_DelOldEntries (database)
  daap_database * database;
{
  int status = 0;
  int count  = 0;
  int i      = 0;
  int leng;
  int date;
  struct stat filestat;
  char name[500];
  daap_song * song;
  
  assert(database != NULL);
  
  if (0 != (status = pthread_mutex_lock(&database->mutex))) {
    fprintf(stderr, "daap_database_DelOldEntries.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  if (database->len == 0 || database->songs == NULL) {
    fprintf(stderr, "daap_database_DelOldEntries.. database is empty. don't bother\n");
    return 0;
  }
  
  for (i = 0; i < database->max; i++) {
    if (NULL == (song = database->songs[i]))
      continue;

    bzero(&name, 500);
    leng = song->assz.valu;
    date = song->asdm.valu;
    bzero(&filestat, sizeof(struct stat));
    status = stat(song->file_path, &filestat);

    memcpy(&name, song->file_path, strlen(song->file_path) > 499 ? 499 : strlen(song->file_path));
    
    // if stat() returns -1, then it was not able to access the file, and thus
    // we will likely not be able to read the file if some poor unsuspecting
    // soul dared try to stream it. as such, there isn't any point in adver-
    // tising its existence, then is there? remove it from the database, and
    // free the memory. be free!
    if (status == -1) {
      fprintf(stderr, "daap_database_DelOldEntries.. DELETE  %s\n", name);
      hash_Remove(database->files, database->songs[i]->file_path, strlen(database->songs[i]->file_path));
      database->songs[i] = NULL;
      database->len--;
      daap_song_Destroy(song);
      count++;
      continue;
    }

    // if the size or last modified time of this file has changed since we 
    // last analyzed it, then it is entirely possible that there are more 
    // changes to be had. reparse the ID3 tags, and mpeg audio frames.
    if (leng != filestat.st_size || date != filestat.st_mtime) {
      fprintf(stderr, "daap_database_DelOldEntries.. UPDATE  %s\n", name);
      daap_database_LoadFile(NULL, name, song);
      count++;
      continue;
    }
    
    fprintf(stderr, "daap_database_DelOldEntries.. CURRENT %s\n", name);
  }
  
  if (0 != (status = pthread_mutex_unlock(&database->mutex))) {
    fprintf(stderr, "daap_database_AppendSong.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return count;
}

daap_int
daap_database_Export (database)
  daap_database * database;
{
  FILE * file;
  int status = 0;
  int i = 0;
  int count = 0;
  daap_song * song;
  buff_t * buff;
  
  daap_int name_song = htonl(0x736f6e67);       // name "song"
  daap_int name_path = htonl(0x70617468);       // name "path"
  daap_int name_name = htonl(0x6e616d65);       // name "name"
  daap_int leng_song = 0;                       // length of song data
  daap_int leng_path = 0;                       // length of song->file_path
  daap_int leng_name = 0;                       // length of song->file_name
  
  assert(database != NULL);
  
  if (0 != (status = pthread_mutex_lock(&database->mutex))) {
    fprintf(stderr, "daap_database_AppendSong.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  if (database->len == 0 || database->songs == NULL) {
    fprintf(stderr, "daap_database_Export.. database is empty. don't bother\n");
    return 0;
  }

  buff = buff_Create();
  buff_Init(buff, 10000);
  
  file = fopen("database.icast", "w");
  assert(file != NULL);
  
  for (i = 0; i < database->max; i++) {
    if (NULL == (song = database->songs[i]))
      continue;

    leng_song = 0;
    buff_ClearAll(buff);
    
    buff_Append(buff, &name_song, 4);
    buff_Append(buff, &leng_song, 4);

    if (song->file_path != NULL) {
      leng_path  = htonl(strlen(song->file_path));
      leng_song += buff_Append(buff, &name_path, 4);
      leng_song += buff_Append(buff, &leng_path, 4);
      leng_song += buff_Append(buff, song->file_path, strlen(song->file_path));
    }
    
    if (song->file_name != NULL) {
      leng_name  = htonl(strlen(song->file_name));
      leng_song += buff_Append(buff, &name_name, 4);
      leng_song += buff_Append(buff, &leng_name, 4);
      leng_song += buff_Append(buff, song->file_name, strlen(song->file_name));
    }
    
    leng_song += daap_Serialize((daap_base *)&song->mikd, buff);
    leng_song += daap_Serialize((daap_base *)&song->miid, buff);
    leng_song += daap_Serialize((daap_base *)&song->minm, buff);
    leng_song += daap_Serialize((daap_base *)&song->mper, buff);
    leng_song += daap_Serialize((daap_base *)&song->asal, buff);
    leng_song += daap_Serialize((daap_base *)&song->agrp, buff);
    leng_song += daap_Serialize((daap_base *)&song->asar, buff);
    leng_song += daap_Serialize((daap_base *)&song->asbr, buff);
    leng_song += daap_Serialize((daap_base *)&song->asbt, buff);
    leng_song += daap_Serialize((daap_base *)&song->ascm, buff);
    leng_song += daap_Serialize((daap_base *)&song->asco, buff);
    leng_song += daap_Serialize((daap_base *)&song->ascp, buff);
    leng_song += daap_Serialize((daap_base *)&song->asda, buff);
    leng_song += daap_Serialize((daap_base *)&song->asdm, buff);
    leng_song += daap_Serialize((daap_base *)&song->asdc, buff);
    leng_song += daap_Serialize((daap_base *)&song->asdn, buff);
    leng_song += daap_Serialize((daap_base *)&song->asdb, buff);
    leng_song += daap_Serialize((daap_base *)&song->aseq, buff);
    leng_song += daap_Serialize((daap_base *)&song->asfm, buff);
    leng_song += daap_Serialize((daap_base *)&song->asgn, buff);
    leng_song += daap_Serialize((daap_base *)&song->asdt, buff);
    leng_song += daap_Serialize((daap_base *)&song->asrv, buff);
    leng_song += daap_Serialize((daap_base *)&song->assr, buff);
    leng_song += daap_Serialize((daap_base *)&song->assz, buff);
    leng_song += daap_Serialize((daap_base *)&song->asst, buff);
    leng_song += daap_Serialize((daap_base *)&song->assp, buff);
    leng_song += daap_Serialize((daap_base *)&song->astm, buff);
    leng_song += daap_Serialize((daap_base *)&song->astc, buff);
    leng_song += daap_Serialize((daap_base *)&song->astn, buff);
    leng_song += daap_Serialize((daap_base *)&song->asur, buff);
    leng_song += daap_Serialize((daap_base *)&song->asyr, buff);
    leng_song += daap_Serialize((daap_base *)&song->asdk, buff);
    leng_song += daap_Serialize((daap_base *)&song->asul, buff);
    leng_song += daap_Serialize((daap_base *)&song->aeNV, buff);
    leng_song += daap_Serialize((daap_base *)&song->aeSP, buff);
    
    // convert the serialize length of the song to network byte order, copy
    // it into the appropriate place in the buffer, and then write it out.
    leng_song = htonl(leng_song);
    memcpy(buff->data + 4, &leng_song, 4);
    fwrite(buff->data, 1, buff->leng, file);
    
    fprintf(stderr, "daap_database_Export.. write %04d bytes for %04d\n", leng_song, song->miid.valu);
  }
  
  fclose(file);
  buff_Destroy(buff);
  
  if (0 != (status = pthread_mutex_unlock(&database->mutex))) {
    fprintf(stderr, "daap_database_AppendSong.. failed to unlock mutex, %d\n", status);
    return -1;
  }

  return count;
}

daap_int
daap_database_AppendSong (database, song)
  daap_database * database;
  daap_song * song;
{
  daap_int songid = -1;
  int status = 0;
  int i = 0;
  
  assert(database != NULL);
  assert(song != NULL);

  if (0 != (status = pthread_mutex_lock(&database->mutex))) {
    fprintf(stderr, "daap_database_AppendSong.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  if (database->len >= database->max - 1) {
    database->max *= 1.5;
    database->songs = (daap_song **)realloc(database->songs, sizeof(daap_song *) * database->max);
    assert(database->songs != NULL);
    fprintf(stderr, "daap_database_AppendSong.. resizing to %d\n", database->max);
  }

  for (i = 0; i < database->max; i++) {
    if (database->songs[i] != NULL)
      continue;
    
    database->songs[i] = song;
    database->len++;
    songid = i;
    break;
  }

  // songid = database->len++;
  // database->songs[songid] = song;
  
  if (0 != (status = pthread_mutex_unlock(&database->mutex))) {
    fprintf(stderr, "daap_database_AppendSong.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return songid;
}

daap_int
daap_database_LoadDir (database, path)
  daap_database * database;
  const char * path;
{
  int count = 0;
  char dirname[255];
  struct dirent * entry;
  DIR * dir;
  
  assert(database != NULL);
  assert(path != NULL);

  dir = opendir(path);
  assert(dir != NULL);
  
  bzero(&dirname, 255);
  memcpy(&dirname, path, strlen(path));
  dirname[strlen(path)] = '/';
  
  while (NULL != (entry = readdir(dir)))
  {
    // ignore files whose name begins with a period (.)
    if (0 == strncmp(entry->d_name, ".", 1))
      continue;

    // copy this new part of the path onto dirname
    bzero((char *)&dirname + strlen(path) + 1, 255 - (strlen(path) + 1));
    memcpy(dirname + strlen(path) + 1, entry->d_name, entry->d_reclen);

    // we found a directory
    if (entry->d_type == DT_DIR) {
      fprintf(stderr, "daap_database_LoadDir.. found a directory, '%s'\n", (char *)&dirname);
      count += daap_database_LoadDir(database, (char *)&dirname);
    }
    // we found a regular file
    else if (entry->d_type == DT_REG) {
      fprintf(stderr, "daap_database_LoadDir.. found a file, '%s'\n", dirname);
      if (NULL != daap_database_LoadFile(database, dirname, NULL))
        count++;
      // fprintf(stderr, "daap_database_LoadDir.. %s: count = %d\n", path, count);
    }
    else
      fprintf(stderr, "daap_database_LoadDir.. unknown file type for %s/%s, %d\n",
              path, entry->d_name, entry->d_type);
  }
  
  closedir(dir);
  
  return count;
}

daap_song *
daap_database_LoadFile (database, fname, song)
  daap_database * database;
  char          * fname;
  daap_song     * song;
{
  int fpos   = 0;
  int songid = 0;
  hash_key_t * key;
  
  daap_song * _song = NULL;
  unsigned int sid = 0;
  FILE * file = NULL;
  struct stat filestat;
  
  assert(fname    != NULL);
  
  // make sure that the file name ends with the ".mp3" extension. granted,
  // this isn't very flexible, and it doesn't take into account capital
  // letters, but, honestly, if the user uses capital extensions, I don't
  // want my program to work for him anyway.
  if (0 != strncmp(fname + (strlen(fname) - 4), ".mp3", 4))
    return NULL;

  if (database != NULL && NULL != (key = hash_Get(database->files, fname, strlen(fname)))) {
    // fprintf(stderr, "daap_database_LoadFile.. file already indexed, %s\n", fname);

    if (song == NULL) {
      // fprintf(stderr, "daap_database_LoadFile.. key->data = 0x%02hhx%02hhx%02hhx%02hhx\n",
      //         key->data[0], key->data[1], key->data[2], key->data[3]);
      memcpy(&sid, key->data, 4);
      _song = database->songs[sid];
      // fprintf(stderr, "daap_database_LoadFile.. _song = %d\n", (int)_song);
    }
    else {
      _song = song;
    }
    
    // try to open the file. return -1 if we fail in this endeavor.
    if (NULL == (file = fopen(fname, "r"))) {
      fprintf(stderr, "daap_database_LoadFile.. could not open file, '%s'\n", fname);
      return NULL;
    }
    
    // fstat() returns all sorts of nifty information about the file in question.
    // particularly, we're interested in the file size (in bytes), and the last
    // modification time (in milliseconds) since Jan 1, 1970.
    fstat(fileno(file), &filestat);

    if (_song->assz.valu == filestat.st_size && _song->asdm.valu == filestat.st_mtime) {
      fclose(file);
      // return _song;
      return NULL;
    }
    
    fclose(file);
  }
  
  if (song == NULL) {
    song = daap_song_Create();
    daap_song_Init(song, 0x6d6c6974);
  }

  if (song->file_path != NULL)
    free(song->file_path);
  
  // copy the full path to this song into the file_path member
  song->file_path = (char *)malloc(strlen(fname) + 1);
  assert(song->file_path != NULL);
  bzero(song->file_path, strlen(fname) + 1);
  memcpy(song->file_path, fname, strlen(fname));
  fprintf(stderr, "daap_database_LoadFile.. [1] song->file_path = %s\n", song->file_path);
  
  if (-1 == (fpos = id3_Load(song, fname))) {
    daap_song_Destroy(song);
    return NULL;
  }

  // start parsing mpeg audio frames starting at the position (fpos)
  // where the ID3 parser left off.
  if (-1 == mpeg_Load(song, fname, fpos)) {
    fprintf(stderr, "daap_database_LoadFile.. mpeg_Load() reported an error\n");
    daap_song_Destroy(song);
    return NULL;
  }
  
  if (song->minm.base.name == 0)
    minm_Init(&song->minm, "Unknown Title");
  
  if (song->asar.base.name == 0)
    asar_Init(&song->asar, "Unknown Artist");
  
  if (song->asal.base.name == 0)
    asal_Init(&song->asal, "Unknown Album");
  
  if (song->asgn.base.name == 0)
    asgn_Init(&song->asgn, "Unknown Genre");

  // if we were passed a database, then assume that we should add
  // this song to that database. if the "user" does not want this,
  // then the user should simply not pass in a database.
  if (database != NULL) {
    songid = daap_database_AppendSong(database, song);
    assert(songid >= 0);
    hash_Put(database->files, song->file_path, strlen(song->file_path), (char *)&songid, 4);
    miid_Init(&song->miid, songid + 1);                       // dmap.itemid
    mper_Init(&song->mper, songid + 1);                       // dmap.persistentid
  }
  
  mikd_Init(&song->mikd, 0x02);                             // dmap.itemkind
  asbt_Init(&song->asbt, 0);                                // daap.songbeatsperminute
  ascm_Init(&song->ascm, "comment");                        // daap.songcomment
  asco_Init(&song->asco, 0);                                // daap.songcompilation
  // ascp_Init(&song->ascp, "composer");                       // daap.songcomposer
  // asda_Init(&song->asda, ...);                              // daap.songdateadded
  asdb_Init(&song->asdb, 0);                                // daap.songdisabled
  aseq_Init(&song->aseq, "");                               // daap.songeqpreset
  asfm_Init(&song->asfm, "mp3");                            // daap.songformat
  asdt_Init(&song->asdt, "description");                    // daap.songdescription
  // asrv_Init(&song->asrv, 0);                                // daap.songrelativevolume
  // asst_Init(&song->asst, ...);                              // daap.songstarttime
  // assp_Init(&song->assp, ...);                              // daap.songstoptime
  
  asdk_Init(&song->asdk, 0x00);                             // daap.songdatakind
  aeNV_Init(&song->aeNV, 0x00);                             // com.apple.itunes.norm-volume
  aeSP_Init(&song->aeSP, 0x00);                             // com.apple.itunes.smart-playlist

  return song;
}

void
daap_database_Destroy (database)
  daap_database * database;
{
  int i = 0;
  
  assert(database != NULL);

  pthread_mutex_destroy(&database->mutex);
  
  for (i = 0; i < database->max; i++)
    if (database->songs[i] != NULL)
      daap_song_Destroy(database->songs[i]);
  
  if (database->songs != NULL)
    free(database->songs);

  hash_Destroy(database->files);
  
  database->len   = 0;
  database->max   = 0;
  database->songs = NULL;
  database->files = NULL;
  
  free(database);
}









#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: VERSION
#endif

inline daap_int
daap_version_Init (__daap_version * daap, daap_int name, daap_short major, daap_short minor)
{
  assert(daap != NULL);
  
  daap->base.name = name;
  daap->base.leng = 4;
  daap->base.type = DAAP_TYPE_INT;
  daap->valu      = (major << 16) | minor;
  
  return 12;
}

inline daap_int
daap_version_Serialize (__daap_version * daap, buff_t * buff)
{
  daap_int name;
  daap_int leng;
  daap_int valu;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  name = htonl(daap->base.name);
  leng = htonl(daap->base.leng);
  valu = htonl(daap->valu);
  
  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  buff_Append(buff, &valu, 4);
  
  return 12;
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: LONG
#endif

inline daap_int
daap_long_Init (__daap_long * daap, daap_int name, daap_long valu)
{
  assert(daap != NULL);

  daap->base.name = name;
  daap->base.leng = 8;
  daap->base.type = DAAP_TYPE_LONG;
  daap->valu      = valu;
  
  return 16;
}

inline daap_int
daap_long_Serialize (__daap_long * daap, buff_t * buff)
{
  daap_int name;
  daap_int leng;
  daap_int valu1;
  daap_int valu2;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  name  = htonl(daap->base.name);
  leng  = htonl(daap->base.leng);
  valu1 = htonl((daap->valu      ) & 0xFFFFFFFF);
  valu2 = htonl((daap->valu >> 32) & 0xFFFFFFFF);

  buff_Append(buff, &name,  4);
  buff_Append(buff, &leng,  4);
  buff_Append(buff, &valu2, 4);
  buff_Append(buff, &valu1, 4);
  
  return 16;
}

inline daap_int
daap_long_Import (daap, name, leng, data)
  __daap_long * daap;
  daap_int name;
  daap_int leng;
  unsigned const char * data;
{
  daap_long valu = 0;
  
  assert(daap != NULL);
  assert(data != NULL);
  assert(leng == 8);
  
  memcpy(&valu, data, 8);
  
  return daap_long_Init(daap, name, valu);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: BYTE
#endif

inline daap_int
daap_byte_Init (__daap_byte * daap, daap_int name, daap_byte valu)
{
  assert(daap != NULL);
  
  daap->base.name = name;
  daap->base.leng = 1;
  daap->base.type = DAAP_TYPE_BYTE;
  daap->valu      = valu;
  
  return 9;
}

inline daap_int
daap_byte_Serialize (__daap_byte * daap, buff_t * buff)
{
  daap_int name;
  daap_int leng;
  
  assert(daap != NULL);
  assert(buff != NULL);

  name = htonl(daap->base.name);
  leng = htonl(daap->base.leng);

  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  buff_Append(buff, &daap->valu, 1);
  
  return 9;
}

inline daap_int
daap_byte_Import (daap, name, leng, data)
  __daap_byte * daap;
  daap_int name;
  daap_int leng;
  unsigned const char * data;
{
  daap_byte valu = 0;
  
  assert(daap != NULL);
  assert(data != NULL);
  assert(leng == 1);

  memcpy(&valu, data, 1);
  
  return daap_byte_Init(daap, name, valu);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: SHORT
#endif

inline daap_int
daap_short_Init (__daap_short * daap, daap_int name, daap_short valu)
{
  assert(daap != NULL);
  
  daap->base.name = name;
  daap->base.leng = 2;
  daap->base.type = DAAP_TYPE_SHORT;
  daap->valu      = valu;
  
  return 10;
}

inline daap_int
daap_short_Serialize (__daap_short * daap, buff_t * buff)
{
  daap_int   name;
  daap_int   leng;
  daap_short valu;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  name = htonl(daap->base.name);
  leng = htonl(daap->base.leng);
  valu = htons(daap->valu);
  
  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  buff_Append(buff, &valu, 2);
  
  return 10;
}

inline daap_int
daap_short_Import (daap, name, leng, data)
  __daap_short * daap;
  daap_int name;
  daap_int leng;
  unsigned const char * data;
{
  daap_short valu = 0;
  
  assert(daap != NULL);
  assert(data != NULL);
  assert(leng == 2);
  
  memcpy(&valu, data, 2);
  
  return daap_short_Init(daap, name, valu);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: DATE
#endif

inline daap_int
daap_date_Init (__daap_date * daap, daap_int name, daap_date valu)
{
  assert(daap != NULL);
  
  daap->base.name = name;
  daap->base.leng = 4;
  daap->base.type = DAAP_TYPE_DATE;
  daap->valu      = valu;
  
  return 12;
}

inline daap_int
daap_date_Serialize (__daap_date * daap, buff_t * buff)
{
  daap_int name;
  daap_int leng;
  daap_int valu;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  name = htonl(daap->base.name);
  leng = htonl(daap->base.leng);
  valu = htonl(daap->valu);
  
  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  buff_Append(buff, &valu, 4);
  
  return 12;
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: INT
#endif

inline daap_int
daap_int_Init (__daap_int * daap, daap_int name, daap_int valu)
{
  assert(daap != NULL);
  
  daap->base.name = name;
  daap->base.leng = 4;
  daap->base.type = DAAP_TYPE_INT;
  daap->valu      = valu;
  
  return 12;
}

inline daap_int
daap_int_Serialize (__daap_int * daap, buff_t * buff)
{
  daap_int name;
  daap_int leng;
  daap_int valu;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  name = htonl(daap->base.name);
  leng = htonl(daap->base.leng);
  valu = htonl(daap->valu);

  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  buff_Append(buff, &valu, 4);
  
  return 12;
}

inline daap_int
daap_int_Import (daap, name, leng, data)
  __daap_int * daap;
  daap_int name;
  daap_int leng;
  unsigned const char * data;
{
  daap_int valu = 0;
  
  assert(daap != NULL);
  assert(data != NULL);
  assert(leng == 4);
  
  memcpy(&valu, data, 4);
  
  return daap_int_Init(daap, name, valu);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: LIST
#endif

inline daap_int
daap_list_Init (__daap_list * daap, daap_int name)
{
  assert(daap != NULL);
  
  daap->base.name = name;
  daap->base.leng = 0;
  daap->base.type = DAAP_TYPE_LIST;
  
  daap->valu.len = 0;
  daap->valu.max = 0;
  
  daap->valu.list = NULL;
  
  return 8;
}

inline daap_int
daap_list_Append (__daap_list * daap, daap_base * valu)
{
  assert(daap != NULL);
  assert(valu != NULL);
  
  if (daap->valu.list == NULL) {
    daap->valu.max  = 5;
    daap->valu.list = (void **)malloc(sizeof(void *) * daap->valu.max);
  }
  else if (daap->valu.max <= daap->valu.len) {
    daap->valu.max *= 1.5;
    daap->valu.list = (void **)realloc(daap->valu.list, sizeof(void *) * daap->valu.max);
  }
  
  assert(daap->valu.list != NULL);
  
  daap->valu.list[daap->valu.len++] = valu;
  
  return daap->valu.len;
}

inline daap_int
daap_list_Serialize (__daap_list * daap, buff_t * buff)
{
  int i = 0;
  int pos = 0;
  daap_int name;
  daap_int leng;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  daap->base.leng = 0;
  
//for (i = 0; i < daap->valu.len; i++)
//  daap->base.leng += 8 + ((daap_base *)daap->valu.list[i])->leng;
  
  name = htonl(daap->base.name);
  leng = 0;
  
  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);

  pos = buff->leng;
  
  for (i = 0; i < daap->valu.len; i++)
    daap->base.leng += daap_Serialize((daap_base *)daap->valu.list[i], buff);

  leng = htonl(daap->base.leng);
  memcpy(buff->data + (pos - 4), &leng, 4);
  
  return 8 + daap->base.leng;
}

inline void
daap_list_Destroy (__daap_list * daap)
{
  int i = 0;
  
  assert(daap != NULL);

  if (daap->valu.list != NULL) {
    for (i = 0; i < daap->valu.len; i++)
      if (daap->valu.list[i] != NULL)
        daap_Destroy((daap_base *)daap->valu.list[i]);
    
    free(daap->valu.list);
  }
  
  daap->base.name = 0;
  daap->base.leng = 0;
  daap->valu.list = NULL;
  daap->valu.len  = 0;
  daap->valu.max  = 0;
  
  free(daap);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Primitive Data Type: STRING
#endif

inline daap_int
daap_string_Init (__daap_string * daap, daap_int name, daap_string valu)
{
  daap_int leng = valu != NULL ? strlen(valu) : 0;
  
  assert(daap != NULL);
  
  if (daap->valu != NULL)
    daap->valu = (char *)realloc(daap->valu, sizeof(char) * leng);
  else
    daap->valu = (char *)malloc(sizeof(char) * leng);
  
  assert(daap->valu != NULL);
  memcpy(daap->valu, valu, leng);
  
  daap->base.name = name;
  daap->base.leng = leng;
  daap->base.type = DAAP_TYPE_STRING;
  
  return 8 + leng;
}

inline daap_int
daap_string_Serialize (__daap_string * daap, buff_t * buff)
{
  daap_int name;
  daap_int leng;
  
  assert(daap != NULL);
  assert(buff != NULL);
  
  name = htonl(daap->base.name);
  leng = htonl(daap->base.leng);

  buff_Append(buff, &name, 4);
  buff_Append(buff, &leng, 4);
  buff_Append(buff, daap->valu, daap->base.leng);
  
  return 8 + daap->base.leng;
}

inline daap_int
daap_string_Import (daap, name, leng, data)
  __daap_string * daap;
  daap_int name;
  daap_int leng;
  unsigned const char * data;
{
  assert(daap != NULL);
  assert(data != NULL);
  
  if (daap->valu != NULL)
    daap->valu = (char *)realloc(daap->valu, sizeof(char) * leng);
  else
    daap->valu = (char *)malloc(sizeof(char) * leng);
  
  assert(daap->valu != NULL);
  memcpy(daap->valu, data, leng);
  
  daap->base.name = name;
  daap->base.leng = leng;
  daap->base.type = DAAP_TYPE_STRING;
  
  return 8 + leng;
}

inline void
daap_string_Destroy (__daap_string * daap)
{
  assert(daap != NULL);
  
  if (daap->valu != NULL)
    free(daap->valu);
  
  daap->base.name = 0;
  daap->base.leng = 0;
  daap->valu      = NULL;
  
  free(daap);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Generic Methods
#endif

daap_int
daap_Serialize (daap_base * daap, buff_t * buff)
{
  assert(daap != NULL);
  assert(buff != NULL);

  switch (daap->type)
  {
    case DAAP_TYPE_BYTE:
      return daap_byte_Serialize((__daap_byte *)daap, buff);
    case DAAP_TYPE_SHORT:
      return daap_short_Serialize((__daap_short *)daap, buff);
    case DAAP_TYPE_INT:
      return daap_int_Serialize((__daap_int *)daap, buff);
    case DAAP_TYPE_LONG:
      return daap_long_Serialize((__daap_long *)daap, buff);
    case DAAP_TYPE_STRING:
      return daap_string_Serialize((__daap_string *)daap, buff);
    case DAAP_TYPE_DATE:
      return daap_date_Serialize((__daap_date *)daap, buff);
    case DAAP_TYPE_VERSION:
      return daap_version_Serialize((__daap_version *)daap, buff);
    case DAAP_TYPE_LIST:
      return daap_list_Serialize((__daap_list *)daap, buff);
    case DAAP_TYPE_SONG:
      return daap_song_Serialize((daap_song *)daap, buff);
    case 0:
      break;
    default:
      fprintf(stderr, "daap_Serialize.. invalid type='%d' name='%d' leng='%d'\n", daap->type, daap->name, daap->leng);
      break;
  }
  
  return 0;
}

daap_int
daap_Import (daap, type, data)
  daap_base * daap;
  daap_short type;
  unsigned const char * data;
{
  daap_int name = 0;
  daap_int leng = 0;
  
  assert(daap != NULL);
  assert(data != NULL);
  
  memcpy(&name, data + 0, 4);
  memcpy(&leng, data + 4, 4);
  
  switch (type)
  {
    case DAAP_TYPE_BYTE:
      return daap_byte_Import   ((__daap_byte     *)daap, name, leng, data + 8);
    case DAAP_TYPE_SHORT:
      return daap_short_Import  ((__daap_short    *)daap, name, leng, data + 8);
    case DAAP_TYPE_INT:
      return daap_int_Import    ((__daap_int      *)daap, name, leng, data + 8);
    case DAAP_TYPE_LONG:
      return daap_long_Import   ((__daap_long     *)daap, name, leng, data + 8);
    case DAAP_TYPE_STRING:
      return daap_string_Import ((__daap_string   *)daap, name, leng, data + 8);
/**
    case DAAP_TYPE_DATE:
      return daap_date_Import   ((__daap_date *)daap,     buff);
    case DAAP_TYPE_VERSION:
      return daap_version_Import((__daap_version *)daap,  buff);
    case DAAP_TYPE_LIST:
      return daap_list_Import   ((__daap_list *)daap,     buff);
    case DAAP_TYPE_SONG:
      return daap_song_Import   ((daap_song *)daap,       buff);
**/
    case 0:
      break;
    default:
      fprintf(stderr, "daap_Import.. invalid type=%d name=%d leng=%d\n", 
              daap->type, daap->name, daap->leng);
      break;
  }
  
  return 0;
}

void
daap_Destroy (daap_base * daap)
{
  assert(daap != NULL);
  
  switch (daap->type)
  {
    case DAAP_TYPE_BYTE:
//    daap_byte_Destroy((__daap_byte *)daap);
      free(daap);
      break;
    case DAAP_TYPE_SHORT:
//    daap_short_Destroy((__daap_short *)daap);
      free(daap);
      break;
    case DAAP_TYPE_INT:
//    daap_int_Destroy((__daap_int *)daap);
      free(daap);
      break;
    case DAAP_TYPE_LONG:
//    daap_long_Destroy((__daap_long *)daap);
      free(daap);
      break;
    case DAAP_TYPE_STRING:
      daap_string_Destroy((__daap_string *)daap);
      break;
    case DAAP_TYPE_DATE:
//    daap_date_Destroy((__daap_date *)daap);
      free(daap);
      break;
    case DAAP_TYPE_VERSION:
//    daap_version_Destroy((__daap_version *)daap);
      free(daap);
      break;
    case DAAP_TYPE_LIST:
      daap_list_Destroy((__daap_list *)daap);
      break;
    default:
      fprintf(stderr, "daap_Destroy.. invalid type, '%d'\n", daap->type);
      break;
  }
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Misc Initializers
#endif

inline daap_int
mstt_Init (daap_mstt * mstt, daap_int valu)
{
  return daap_int_Init((__daap_int *)mstt, 0x6d737474, valu);
}

inline daap_int
mpro_Init (daap_mpro * mpro, daap_short major, daap_short minor)
{
  return daap_version_Init((__daap_version *)mpro, 0x6d70726f, major, minor);
}

inline daap_int
apro_Init (daap_apro * apro, daap_short major, daap_short minor)
{
  return daap_version_Init((__daap_version *)apro, 0x6170726f, major, minor);
}

inline daap_int
msix_Init (daap_msix * msix, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msix, 0x6d736978, valu);
}

inline daap_int
msex_Init (daap_msex * msex, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msex, 0x6d736578, valu);
}

inline daap_int
msup_Init (daap_msup * msup, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msup, 0x6d737570, valu);
}

inline daap_int
msal_Init (daap_msal * msal, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msal, 0x6d73616c, valu);
}

inline daap_int
mstm_Init (daap_mstm * mstm, daap_int valu)
{
  return daap_int_Init((__daap_int *)mstm, 0x6d73746d, valu);
}

inline daap_int
mslr_Init (daap_mslr * mslr, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)mslr, 0x6d736c72, valu);
}

inline daap_int
msqy_Init (daap_msqy * msqy, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msqy, 0x6d737179, valu);
}

inline daap_int
minm_Init (daap_minm * minm, daap_string valu)
{
  return daap_string_Init((__daap_string *)minm, 0x6d696e6d, valu);
}

inline daap_int
msrs_Init (daap_msrs * msrs, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msrs, 0x6d737273, valu);
}

inline daap_int
msbr_Init (daap_msbr * msbr, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)msbr, 0x6d736272, valu);
}

inline daap_int
mspi_Init (daap_mspi * mspi, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)mspi, 0x6d737069, valu);
}

inline daap_int
msdc_Init (daap_msdc * msdc, daap_int valu)
{
  return daap_int_Init((__daap_int *)msdc, 0x6d736463, valu);
}

inline daap_int
mdcl_Init (daap_mdcl * mdcl)
{
  return daap_list_Init((__daap_list *)mdcl, 0x6d64636c);
}

inline daap_int
mcna_Init (daap_mcna * mcna, daap_string valu)
{
  return daap_string_Init((__daap_string *)mcna, 0x6d636e61, valu);
}

inline daap_int
mcnm_Init (daap_mcnm * mcnm, daap_int valu)
{
  return daap_int_Init((__daap_int *)mcnm, 0x6d636e6d, valu);
}

inline daap_int
mcty_Init (daap_mcty * mcty, daap_short valu)
{
  return daap_short_Init((__daap_short *)mcty, 0x6d637479, valu);
}

inline daap_int
mlid_Init (daap_mlid * mlid, daap_int valu)
{
  return daap_int_Init((__daap_int *)mlid, 0x6d6c6964, valu);
}

inline daap_int
musr_Init (daap_musr * musr, daap_int valu)
{
  return daap_int_Init((__daap_int *)musr, 0x6d757372, valu);
}

inline daap_int
muty_Init (daap_muty * muty, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)muty, 0x6d757479, valu);
}

inline daap_int
mtco_Init (daap_mtco * mtco, daap_int valu)
{
  return daap_int_Init((__daap_int *)mtco, 0x6d74636f, valu);
}

inline daap_int
mrco_Init (daap_mrco * mrco, daap_int valu)
{
  return daap_int_Init((__daap_int *)mrco, 0x6d72636f, valu);
}

inline daap_int
mlcl_Init (daap_mlcl * mlcl)
{
  return daap_list_Init((__daap_list *)mlcl, 0x6d6c636c);
}

inline daap_int
mlit_Init (daap_mlit * mlit)
{
  return daap_list_Init((__daap_list *)mlit, 0x6d6c6974);
}

inline daap_int
miid_Init (daap_miid * miid, daap_int valu)
{
  return daap_int_Init((__daap_int *)miid, 0x6d696964, valu);
}

inline daap_int
mper_Init (daap_mper * mper, daap_long valu)
{
  return daap_long_Init((__daap_long *)mper, 0x6d706572, valu);
}

inline daap_int
mimc_Init (daap_mimc * mimc, daap_int valu)
{
  return daap_int_Init((__daap_int *)mimc, 0x6d696d63, valu);
}

inline daap_int
mctc_Init (daap_mctc * mctc, daap_int valu)
{
  return daap_int_Init((__daap_int *)mctc, 0x6d637463, valu);
}

inline daap_int
asal_Init (asal, valu)
  daap_asal * asal;
  daap_string valu;
{
  return daap_string_Init((__daap_string *)asal, 0x6173616c, valu);
}

inline daap_int
asbr_Init (asbr, valu)
  daap_asbr * asbr;
  daap_short valu;
{
  return daap_short_Init((__daap_short *)asbr, 0x61736272, valu);
}

inline daap_int
asfm_Init (asfm, valu)
  daap_asfm * asfm;
  daap_string valu;
{
  return daap_string_Init((__daap_string *)asfm, 0x6173666d, valu);
}

inline daap_int
mikd_Init (daap_mikd * mikd, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)mikd, 0x6d696b64, valu);
}

inline daap_int
assz_Init (daap_assz * assz, daap_int valu)
{
  return daap_int_Init((__daap_int *)assz, 0x6173737a, valu);
}

inline daap_int
asdk_Init (daap_asdk * asdk, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)asdk, 0x6173646b, valu);
}

inline daap_int
assr_Init (daap_assr * assr, daap_int valu)
{
  return daap_int_Init((__daap_int *)assr, 0x61737372, valu);
}

inline daap_int
aeNV_Init (daap_aeNV * aeNV, daap_int valu)
{
  return daap_int_Init((__daap_int *)aeNV, 0x61654e56, valu);
}

inline daap_int
aeSP_Init (daap_aeSP * aeSP, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)aeSP, 0x61655350, valu);
}

inline daap_int
asar_Init (daap_asar * asar, daap_string valu)
{
  return daap_string_Init((__daap_string *)asar, 0x61736172, valu);
}

inline daap_int
asbt_Init (daap_asbt * asbt, daap_short valu)
{
  return daap_short_Init((__daap_short *)asbt, 0x61736274, valu);
}

inline daap_int
ascm_Init (daap_ascm * ascm, daap_string valu)
{
  return daap_string_Init((__daap_string *)ascm, 0x6173636d, valu);
}

inline daap_int
asco_Init (daap_asco * asco, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)asco, 0x6173636f, valu);
}

inline daap_int
ascp_Init (daap_ascp * ascp, daap_string valu)
{
  return daap_string_Init((__daap_string *)ascp, 0x61736370, valu);
}

inline daap_int
asdc_Init (daap_asdc * asdc, daap_short valu)
{
  return daap_short_Init((__daap_short *)asdc, 0x61736463, valu);
}

inline daap_int
asdn_Init (daap_asdn * asdn, daap_short valu)
{
  return daap_short_Init((__daap_short *)asdn, 0x6173646e, valu);
}

inline daap_int
asdb_Init (daap_asdb * asdb, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)asdb, 0x61736462, valu);
}

inline daap_int
aseq_Init (daap_aseq * aseq, daap_string valu)
{
  return daap_string_Init((__daap_string *)aseq, 0x61736571, valu);
}

inline daap_int
asgn_Init (daap_asgn * asgn, daap_string valu)
{
  return daap_string_Init((__daap_string *)asgn, 0x6173676e, valu);
}

inline daap_int
asdt_Init (daap_asdt * asdt, daap_string valu)
{
  return daap_string_Init((__daap_string *)asdt, 0x61736474, valu);
}

inline daap_int
asrv_Init (daap_asrv * asrv, daap_byte valu)
{
  return daap_byte_Init((__daap_byte *)asrv, 0x61737276, valu);
}

inline daap_int
mcti_Init (daap_mcti * mcti, daap_int valu)
{
  return daap_int_Init((__daap_int *)mcti, 0x6d637469, valu);
}

inline daap_int
asyr_Init (daap_asyr * asyr, daap_short valu)
{
  return daap_short_Init((__daap_short *)asyr, 0x61737972, valu);
}

inline daap_int
astc_Init (daap_astc * astc, daap_short valu)
{
  return daap_short_Init((__daap_short *)astc, 0x61737463, valu);
}

inline daap_int
astn_Init (daap_astn * astn, daap_short valu)
{
  return daap_short_Init((__daap_short *)astn, 0x6173746e, valu);
}

inline daap_int
astm_Init (daap_astm * astm, daap_int valu)
{
  return daap_int_Init((__daap_int *)astm, 0x6173746d, valu);
  
}

inline daap_int
asdm_Init (daap_asdm * asdm, daap_date valu)
{
  return daap_date_Init((__daap_date *)asdm, 0x6173646d, valu);
}
