#ifndef __DAAP_H
#define __DAAP_H

#include <time.h>         // time()

#include "buff.h"
#include "hash.h"

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** DAAP Data Types
#endif

#define DAAP_TYPE_BYTE    0x0001
#define DAAP_TYPE_SHORT   0x0003
#define DAAP_TYPE_INT     0x0005
#define DAAP_TYPE_LONG    0x0007
#define DAAP_TYPE_STRING  0x0009
#define DAAP_TYPE_DATE    0x000A
#define DAAP_TYPE_VERSION 0x000B
#define DAAP_TYPE_LIST    0x000C
#define DAAP_TYPE_SONG    0x0101

// the following typedefs define the eight daap primitive data types
// these are used in forming the more complex data types and structures
// used by daap.
//
typedef unsigned char       daap_byte;
typedef unsigned short      daap_short;
typedef unsigned int        daap_int;
typedef unsigned long long  daap_long;
typedef char *              daap_string;
typedef unsigned int        daap_date;
typedef unsigned int        daap_version;

typedef struct
{
  int len;
  int max;
  void ** list;
} daap_list;

// the daap protocol specifies a four byte int for the value name, a
// four byte int for the value length, followed immediately be the
// value itself. this daap_base type provides the first two fields,
// which are common to _every_ daap data type.
//
typedef struct
{
  daap_int name;
  daap_int leng;
  daap_short type;
} daap_base;

typedef struct { daap_base base; daap_byte    valu; } __daap_byte;
typedef struct { daap_base base; daap_short   valu; } __daap_short;
typedef struct { daap_base base; daap_int     valu; } __daap_int;
typedef struct { daap_base base; daap_long    valu; } __daap_long;
typedef struct { daap_base base; daap_string  valu; } __daap_string;
typedef struct { daap_base base; daap_date    valu; } __daap_date;
typedef struct { daap_base base; daap_version valu; } __daap_version;
typedef struct { daap_base base; daap_list    valu; } __daap_list;

// the following eighty-five typedefs represent the data types/fields/whatever
// used by the daap protocol. these types are all derived from the eight more
// primitive types defined above. every type struct starts with daap_bas, which
// provides the name and length fields which are required by every daap type.
//
typedef __daap_int      daap_miid;          // dmap.itemid
typedef __daap_string   daap_minm;          // dmap.itemname
typedef __daap_byte     daap_mikd;          // dmap.itemkind
typedef __daap_long     daap_mper;          // dmap.persistentid
typedef __daap_list     daap_mcon;          // dmap.container
typedef __daap_int      daap_mcti;          // dmap.containeritemid
typedef __daap_int      daap_mpco;          // dmap.parentcontainerid
typedef __daap_int      daap_mstt;          // dmap.status
typedef __daap_string   daap_msts;          // dmap.statusstring
typedef __daap_int      daap_mimc;          // dmap.itemcount
typedef __daap_int      daap_mctc;          // dmap.containercount
typedef __daap_int      daap_mrco;          // dmap.returnedcount
typedef __daap_int      daap_mtco;          // dmap.specifiedtotalcount
typedef __daap_list     daap_mlcl;          // dmap.listing
typedef __daap_list     daap_mlit;          // dmap.listingitem
typedef __daap_list     daap_mbcl;          // dmap.bag
typedef __daap_list     daap_mdcl;          // dmap.dictionary
typedef __daap_byte     daap_msau;          // dmap.authenticationmethod
typedef __daap_byte     daap_mslr;          // dmap.loginrequired
typedef __daap_version  daap_mpro;          // dmap.protocolversion
typedef __daap_byte     daap_msal;          // dmap.supportsautologout
typedef __daap_byte     daap_msup;          // dmap.supportsupdate
typedef __daap_byte     daap_mspi;          // dmap.supportspersistentids
typedef __daap_byte     daap_msex;          // dmap.supportsextensions
typedef __daap_byte     daap_msbr;          // dmap.supportsbrowse
typedef __daap_byte     daap_msqy;          // dmap.supportsquery
typedef __daap_byte     daap_msix;          // dmap.supportsindex
typedef __daap_byte     daap_msrs;          // dmap.supportsresolve
typedef __daap_int      daap_mstm;          // dmap.timeoutinterval
typedef __daap_int      daap_msdc;          // dmap.databasescount
typedef __daap_int      daap_mlid;          // dmap.sessionid
typedef __daap_int      daap_musr;          // dmap.serverrevision
typedef __daap_byte     daap_muty;          // dmap.updatetype
typedef __daap_list     daap_mudl;          // dmap.deletedidlisting
typedef __daap_int      daap_mcnm;          // dmap.contentcodesnumber
typedef __daap_string   daap_mcna;          // dmap.contentcodesname
typedef __daap_short    daap_mcty;          // dmap.contentcodestype
typedef __daap_version  daap_apro;          // daap.protocolversion
typedef __daap_list     daap_abro;          // daap.databasebrowse
typedef __daap_list     daap_abal;          // daap.browsealbumlisting
typedef __daap_list     daap_abar;          // daap.browseartistlisting
typedef __daap_list     daap_abcp;          // daap.browsecomposerlisting
typedef __daap_list     daap_abgn;          // daap.browsegenrelisting
typedef __daap_string   daap_asal;          // daap.songalbum
typedef __daap_string   daap_agrp;          // daap.songgrouping
typedef __daap_string   daap_asar;          // daap.songartist
typedef __daap_short    daap_asbt;          // daap.songbeatsperminute
typedef __daap_short    daap_asbr;          // daap.songbitrate
typedef __daap_string   daap_ascm;          // daap.songcomment
typedef __daap_byte     daap_asco;          // daap.songcompilation
typedef __daap_string   daap_ascp;          // daap.songcomposer
typedef __daap_date     daap_asda;          // daap.songdateadded
typedef __daap_date     daap_asdm;          // daap.songdatemodified
typedef __daap_short    daap_asdc;          // daap.songdisccount
typedef __daap_short    daap_asdn;          // daap.songdiscnumber
typedef __daap_byte     daap_asdb;          // daap.songdisabled
typedef __daap_string   daap_aseq;          // daap.songeqpreset
typedef __daap_string   daap_asfm;          // daap.songformat
typedef __daap_string   daap_asgn;          // daap.songgenre
typedef __daap_string   daap_asdt;          // daap.songdescription
typedef __daap_byte     daap_asrv;          // daap.songrelativevolume
typedef __daap_int      daap_assr;          // daap.songsamplerate
typedef __daap_int      daap_assz;          // daap.songsize
typedef __daap_int      daap_asst;          // daap.songstarttime
typedef __daap_int      daap_assp;          // daap.songstoptime
typedef __daap_int      daap_astm;          // daap.songtime
typedef __daap_short    daap_astc;          // daap.songtrackcount
typedef __daap_short    daap_astn;          // daap.songtracknumber
typedef __daap_byte     daap_asur;          // daap.songuserrating
typedef __daap_short    daap_asyr;          // daap.songyear
typedef __daap_byte     daap_asdk;          // daap.songdatakind
typedef __daap_string   daap_asul;          // daap.songdataurl
typedef __daap_byte     daap_abpl;          // daap.baseplaylist
typedef __daap_list     daap_arsv;          // daap.resolve
typedef __daap_list     daap_arif;          // daap.resolveinfo
typedef __daap_int      daap_aeNV;          // com.apple.itunes.norm-volume
typedef __daap_byte     daap_aeSP;          // com.apple.itunes.smart-playlist

// this msrv structure is used as the basis for the /server-info request, which is
// typically the first thing an iTunes client will request. it provides basic
// information about the server, it's features, protocols, and security stuff.
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_mpro mpro;               // dmap protocol version;
  daap_apro apro;               // daap protocol
  daap_msix msix;               // does the server support indexing?
  daap_msex msex;               // does the server support extensions?
  daap_msup msup;               // does the server support update?
  daap_msal msal;               // does the server support auto logout?
  daap_mstm mstm;               // timeout interval
  daap_mslr mslr;               // is login required?
  daap_msqy msqy;               // does the server support queries?
  daap_minm minm;               // server name
  daap_msrs msrs;               // does the server support resolve?
  daap_msbr msbr;               // does the server support browsing?
  daap_mspi mspi;               // does the server support persistent ids?
  daap_msdc msdc;               // number of databases
} daap_msrv;

// REQUEST: /content-codes -- a list of all of the content codes the server supports
//
// this mccr structure is used as the basis for the /content-codes request, which is
// typically the second thing an iTunes client will request. a server sends this
// fairly lengthy reply, with one mdcl defined for each content code that the server
// supports (eighty-five, presently). each content code is specified by its four
// letter name (mcnm), and it's "package path" (mcna) and it's associated primitive
// data type (mcty).
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_list mdcl;               // dictionary entry (one for each content code)
} daap_mccr;

// REQUEST: /login -- a request to authenticate and get a session id
//
// this mlog structure is used as the basis for the /login request, which is typically
// the third thing an iTunes client will request. if login is required, then the cgi
// request from iTunes will provide a user (blank) and a password. if this password is
// valid, or if no login is required, a session id is created and returned to the client.
// this session is used with all subsequent requests from the client. however, even with
// this session id, if login is required, each request from the client must provide the
// base64 encoded user name and password.
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_mlid mlid;               // session id
} daap_mlog;

// REQUEST: /server/update? -- first time it wants the server status, and subsequent
//                             calls it wants to be informed of updates
//
// this mupd structure is used as the basis for the initial /server/update? request,
// which is typically the fourth thing an iTunes client will request. the first time
// this request is received for a particular session, it is a request to get all of
// the initial data ... which doesn't seem to be much. the subsequent request by this
// name is to catch updates to the server music database.
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_musr musr;               // the server revision
} daap_mupd;

// REQUEST: /databases? -- list of all databases on the server
//
// this avdb structure is used as the basis for the /databases? request, which is
// typically the fifth thing an iTunes client will request. an mlit is present in
// the mlcl structure for each available database. presently, iTunes seems to only
// offer one database - being the comprehensive list of all the music it knows of.
//
// the mlit structure has the following fields: miid, mper, minm, mimc, and mctc
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_muty muty;               // update type - always 0 for new
  daap_mtco mtco;               // total number of matching records
  daap_mrco mrco;               // total number of records returned
  daap_mlcl mlcl;               // listing of records
} daap_avdb;

// REQUEST: /databases/<dbid>/items? -- list of all songs in the given database
//
// this adbs structure is used as the basis for the /databases/<dbid>/items? request
// which is typically the sixth thing an iTunes client will request. an mlit is present
// in the mlcl structure for each song available in the database. as there is typically
// only one database that is available, ever, this should effectively be a comprehensive
// list of all music.
//
// the mlit structure contains many fields for each song. one important field to remember
// is the miid field, which is used as the unique-id on which the playlist "links" to
// list the songs present in the playlist.
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_muty muty;               // update type (0)
  daap_mtco mtco;               // matched items
  daap_mrco mrco;               // items in message
  daap_mlcl mlcl;               // list of items
} daap_adbs;

// REQUEST: /databases/<dbid>/containers? -- list of all playlists in the given database
//
// this aply structure is used as the basis for the /databases/<dbid>/containers? request
// which is typically the seventh thing an iTunes client will request. an mlit is present
// in the mlcl structure for each playlist available in the given database. again, as there
// is typically only one database ever available, this should be a complete list of all
// playlists available.
//
// the mlit structure has the following fields: miid, mper, minm, mimc, and aeSP
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_muty muty;               // update type (0)
  daap_mtco mtco;               // matched items
  daap_mrco mrco;               // items in message
  daap_mlcl mlcl;               // list of items
} daap_aply;

// REQUEST: /databases/<dbid>/containers/<plid>/items? -- list of songs in the given playlist
//
// this apso structure is used as the basis for the /databases/<dbid>/containers/<plid>/items?
// request which is typically the eighth thing an iTunes client will request. an mlit is
// present in the mlcl structure for each song in the given playlist.
//
// the mlit structure has the following fields: mikd, miid, and mcti. the miid is the important
// value which links this song reference to the actual song, as provided in the full song
// dump.
//
typedef struct
{
  daap_base base;               // name and length
  daap_mstt mstt;               // status (200)
  daap_muty muty;               // update type (0)
  daap_mtco mtco;               // matched items
  daap_mrco mrco;               // items in message
  daap_mlcl mlcl;               // list of items
} daap_apso;

typedef struct
{
  unsigned short daap_type;
  char * daap_name;
  char * cname;
} data_type;

typedef struct
{
  char * name;        // mstt, muty, mtco, etc
  char * path;        // dmap.itemname, dmap.itemkind, etc
  short  type;        // DAAP_TYPE_BYTE, DAAP_TYPE_LIST, etc
} daap_type;

typedef struct
{
  time_t timeout;             // time at which session has expired
  time_t created;             // time at which session was created
  unsigned short session;     // two-byte session id for this client
  int trans_count;            // session transaction count
  int song_count;             // number of songs streamed
} daap_session;

typedef struct
{
  daap_base base;             // name and length
  
  char * file_name;           // file name, eg, song.mp3
  char * file_path;           // file path, eg, ./music/song.mp3
  
  daap_miid miid;             // dmap.itemid
  daap_minm minm;             // dmap.itemname
  daap_mikd mikd;             // dmap.itemkind
  daap_mper mper;             // dmap.persistentid
  daap_asal asal;             // daap.songalbum
  daap_agrp agrp;             // daap.songgroup
  daap_asar asar;             // daap.songartist
  daap_asbr asbr;             // daap.songbitrate
  daap_asbt asbt;             // daap.songbeatsperminute
  daap_ascm ascm;             // daap.songcomment
  daap_asco asco;             // daap.songcompilation
  daap_ascp ascp;             // daap.songcomposer
  daap_asda asda;             // daap.songdateadded
  daap_asdm asdm;             // daap.songdatemodified
  daap_asdc asdc;             // daap.songdisccount
  daap_asdn asdn;             // daap.songdiscnumber
  daap_asdb asdb;             // daap.songdisabled
  daap_aseq aseq;             // daap.songeqpreset
  daap_asfm asfm;             // daap.songformat
  daap_asgn asgn;             // daap.songgenre
  daap_asdt asdt;             // daap.songdescription
  daap_asrv asrv;             // daap.songrelativevolume
  daap_assr assr;             // daap.songsamplerate
  daap_assz assz;             // daap.songsize
  daap_asst asst;             // daap.songstarttime
  daap_assp assp;             // daap.songstoptime
  daap_astm astm;             // daap.songtime
  daap_astc astc;             // daap.songtrackcount
  daap_astn astn;             // daap.songtracknumber
  daap_asur asur;             // daap.songuserrating
  daap_asyr asyr;             // daap.songyear
  daap_asdk asdk;             // daap.songdatakind
  daap_asul asul;             // daap.songdataurl
  daap_aeNV aeNV;             // com.apple.itunes.norm-volume
  daap_aeSP aeSP;             // com.apple.itunes.smart-playlist
} daap_song;

typedef struct
{
  pthread_mutex_t mutex;      // mutex for thread-safe operations
  int len;                    // number of entries in database
  int max;                    // present size of database (malloc)
  hash_t * files;             // keyed on file names; values are points to daap_song's
  daap_song ** songs;         // list of songs
} daap_database;

typedef struct
{
  char * name;
  
  daap_miid miid;             // dmap.itemid
  daap_mper mper;             // dmap.persistentid
  daap_minm minm;             // dmap.itemname
  daap_mimc mimc;             // dmap.itemcount
  daap_aeSP aeSP;             // com.apple.itunes.smart-playlist
} daap_playlist;

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** Function Prototypes
#endif

//
// the following functions take as arguments some basic C data types, and some
// other necessary information, and write to the 'data' a DAAP representation of
// said data. each function returns the length of the data written, in bytes
//
// static inline int WriteDAAPLong     (char * data, char * name, daap_long valu);
// static inline int WriteDAAPInt      (char * data, char * name, daap_int valu);
// static inline int WriteDAAPShort    (char * data, char * name, daap_short valu);
// static inline int WriteDAAPByte     (char * data, char * name, daap_byte valu);
// static inline int WriteDAAPVersion  (char * data, char * name, daap_short major, daap_short minor);
// static inline int WriteDAAPList     (char * data, char * name, char * valu, daap_int leng);
// static inline int WriteDAAPString   (char * data, char * name, daap_string valu);
// static inline int WriteDAAPMdcl     (char * data, char * numb, char * name, daap_short type);
//

// 01 -- daap://server/server-info
daap_msrv * msrv_Create ();
void msrv_Init (daap_msrv *);
buff_t * msrv_Serialize (daap_msrv *);
void msrv_Destroy (daap_msrv *);

// 02 -- daap://server/content-codes
daap_mccr * mccr_Create ();
void mccr_Init (daap_mccr *);
void mccr_AddContentCode (daap_mccr *, daap_mdcl *);
buff_t * mccr_Serialize (daap_mccr *);
void mccr_Destroy (daap_mccr *);

// 03 -- daap://server/login
daap_mlog * mlog_Create ();
void mlog_Init (daap_mlog *);
buff_t * mlog_Serialize (daap_mlog *);
void mlog_Destroy (daap_mlog *);

// 04 -- daap://server/update
daap_mupd * mupd_Create ();
void mupd_Init (daap_mupd *);
buff_t * mupd_Serialize (daap_mupd *);
void mupd_Destroy (daap_mupd *);

// 05 -- daap://server/databases
daap_avdb * avdb_Create ();
void avdb_Init (daap_avdb *);
void avdb_AddRecord (daap_avdb *, daap_mlit *);
buff_t * avdb_Serialize (daap_avdb *);
void avdb_Destroy (daap_avdb *);

// 06 -- daap://server/databases/<dbid>/items
daap_adbs * adbs_Create ();
void adbs_Init (daap_adbs *);
void adbs_AddSong (daap_adbs *, daap_song *);
void adbs_AddRecord (daap_adbs *, daap_mlit * mlit);
buff_t * adbs_Serialize(daap_adbs *);
void adbs_Destroy (daap_adbs *);

// 07-- daap://server/databases/<dbid>/containers
daap_aply * aply_Create ();
void aply_Init (daap_aply *);
void aply_AddPlaylist (daap_aply *, daap_playlist *);
void aply_AddRecord (daap_aply *, daap_mlit * mlit);
buff_t * aply_Serialize(daap_aply *);
void aply_Destroy (daap_aply *);

// 08 -- daap://server/databases/<dbid>/containers/<plid>/items
daap_apso * apso_Create();
void apso_Init(daap_apso *);
void apso_AddSong (daap_apso *, daap_song *);
void apso_AddRecord (daap_apso *, daap_mlit *);
buff_t * apso_Serialize(daap_apso *);
void apso_Destroy (daap_apso *);

inline daap_int daap_version_Init (__daap_version *, daap_int name, daap_short major, daap_short minor);
inline daap_int daap_version_Serialize(__daap_version *, buff_t *);
inline daap_int daap_long_Init (__daap_long *, daap_int name, daap_long valu);
inline daap_int daap_long_Serialize (__daap_long *, buff_t *);
inline daap_int daap_long_Import (__daap_long *, daap_int name, daap_int leng, unsigned const char * data);
inline daap_int daap_byte_Init (__daap_byte *, daap_int name, daap_byte valu);
inline daap_int daap_byte_Serialize (__daap_byte *, buff_t *);
inline daap_int daap_byte_Import (__daap_byte *, daap_int name, daap_int leng, unsigned const char * data);
inline daap_int daap_short_Init (__daap_short *, daap_int name, daap_short valu);
inline daap_int daap_short_Serialize (__daap_short *, buff_t *);
inline daap_int daap_short_Import (__daap_short *, daap_int name, daap_int leng, unsigned const char * data);
inline daap_int daap_date_Init (__daap_date *, daap_int name, daap_date valu);
inline daap_int daap_date_Serialize (__daap_date *, buff_t *);
inline daap_int daap_int_Init (__daap_int *, daap_int name, daap_int valu);
inline daap_int daap_int_Serialize (__daap_int *, buff_t *);
inline daap_int daap_int_Import (__daap_int *, daap_int name, daap_int leng, unsigned const char * data);
inline daap_int daap_list_Init (__daap_list *, daap_int name);
inline daap_int daap_list_Append (__daap_list *, daap_base * valu);
inline daap_int daap_list_Serialize (__daap_list *, buff_t *);
inline void daap_list_Destroy (__daap_list *);
inline daap_int daap_string_Init (__daap_string *, daap_int name, daap_string valu);
inline daap_int daap_string_Serialize (__daap_string *, buff_t *);
inline daap_int daap_string_Import (__daap_string *, daap_int name, daap_int leng, unsigned const char * data);
inline void daap_string_Destroy (__daap_string *);

daap_int daap_Serialize (daap_base * daap, buff_t * buff);
daap_int daap_Import (daap_base * daap, daap_short type, unsigned const char * data);
void daap_Destroy (daap_base * daap);

inline daap_int mstt_Init (daap_mstt *, daap_int valu);
inline daap_int msix_Init (daap_msix *, daap_byte valu);
inline daap_int msex_Init (daap_msex *, daap_byte valu);
inline daap_int msup_Init (daap_msup *, daap_byte valu);
inline daap_int msal_Init (daap_msal *, daap_byte valu);
inline daap_int mstm_Init (daap_mstm *, daap_int valu);
inline daap_int mslr_Init (daap_mslr *, daap_byte valu);
inline daap_int msqy_Init (daap_msqy *, daap_byte valu);
inline daap_int minm_Init (daap_minm *, daap_string valu);
inline daap_int msrs_Init (daap_msrs *, daap_byte valu);
inline daap_int msbr_Init (daap_msbr *, daap_byte valu);
inline daap_int mspi_Init (daap_mspi *, daap_byte valu);
inline daap_int apro_Init (daap_apro *, daap_short major, daap_short minor);
inline daap_int mpro_Init (daap_mpro *, daap_short major, daap_short minor);
inline daap_int mdcl_Init (daap_mdcl *);
inline daap_int mcna_Init (daap_mcna *, daap_string valu);
inline daap_int mcnm_Init (daap_mcnm *, daap_int valu);
inline daap_int mcty_Init (daap_mcty *, daap_short valu);
inline daap_int mlid_Init (daap_mlid *, daap_int valu);
inline daap_int mtco_Init (daap_mtco *, daap_int valu);
inline daap_int mrco_Init (daap_mrco *, daap_int valu);

inline daap_int musr_Init (daap_musr *, daap_int valu);

inline daap_int msdc_Init (daap_msdc *, daap_int valu);

inline daap_int muty_Init (daap_muty *, daap_byte valu);
inline daap_int mlcl_Init (daap_mlcl *);
inline daap_int mlit_Init (daap_mlit *);
inline daap_int miid_Init (daap_miid *, daap_int valu);
inline daap_int mper_Init (daap_mper *, daap_long valu);
inline daap_int mimc_Init (daap_mimc *, daap_int valu);
inline daap_int mctc_Init (daap_mctc *, daap_int valu);

inline daap_int asal_Init (daap_asal *, daap_string valu);
inline daap_int asbr_Init (daap_asbr *, daap_short valu);
inline daap_int asfm_Init (daap_asfm *, daap_string valu);
inline daap_int mikd_Init (daap_mikd *, daap_byte valu);
inline daap_int assz_Init (daap_assz *, daap_int valu);
inline daap_int asdk_Init (daap_asdk *, daap_byte valu);
inline daap_int assr_Init (daap_assr *, daap_int valu);
inline daap_int aeNV_Init (daap_aeNV *, daap_int valu);
inline daap_int aeSP_Init (daap_aeSP *, daap_byte valu);

inline daap_int asar_Init (daap_asar *, daap_string valu);
inline daap_int asbt_Init (daap_asbt *, daap_short valu);
inline daap_int ascm_Init (daap_ascm *, daap_string valu);
inline daap_int asco_Init (daap_asco *, daap_byte valu);
inline daap_int ascp_Init (daap_ascp *, daap_string valu);
inline daap_int asdc_Init (daap_asdc *, daap_short valu);
inline daap_int asdn_Init (daap_asdn *, daap_short valu);
inline daap_int asdb_Init (daap_asdb *, daap_byte valu);
inline daap_int aseq_Init (daap_aseq *, daap_string valu);
inline daap_int asgn_Init (daap_asgn *, daap_string valu);
inline daap_int asdt_Init (daap_asdt *, daap_string valu);
inline daap_int asrv_Init (daap_asrv *, daap_byte valu);
inline daap_int asyr_Init (daap_asyr *, daap_short valu);
inline daap_int astc_Init (daap_astc *, daap_short valu);
inline daap_int astn_Init (daap_astn *, daap_short valu);
inline daap_int astm_Init (daap_astm *, daap_int valu);
// inline daap_int asda_Init (daap_asda *, daap_date valu);
inline daap_int asdm_Init (daap_asdm *, daap_date valu);

inline daap_int mcti_Init (daap_mcti *, daap_int valu);


daap_song * daap_song_Create ();
void daap_song_Init (daap_song *, daap_int name);
void daap_song_Destroy (daap_song *);

// daap_database kDefaultDatabase;
// daap_adbs kDefaultAdbs;
// daap_apso kDefaultApso;

daap_database * daap_database_Create ();
void daap_database_Init (daap_database *);
daap_int daap_database_Import (daap_database *);
daap_int daap_database_Export (daap_database *);
daap_int daap_database_Update (daap_database *, const char * path);
daap_int daap_database_DelOldEntries (daap_database *);
daap_int daap_database_LoadDir (daap_database *, const char * path);
daap_song * daap_database_LoadFile (daap_database *, char * fname, daap_song * song);
daap_int daap_database_AppendSong (daap_database *, daap_song *);
void daap_database_Destroy (daap_database *);

#endif /* __DAAP_H */
