#include <stdlib.h>         // malloc, realloc, etc
#include <assert.h>         // assert
#include <stdio.h>          // fprintf
#include <strings.h>        // bzero
#include <unistd.h>         // close
#include <sys/types.h>      // socket stuff
#include <sys/socket.h>     // socket stuff
#include <sys/select.h>     // select stuff
#include <sys/time.h>       // struct timeval
#include <fcntl.h>          // fcntl
#include <zlib.h>           // compress()

#include "http.h"
#include "inet.h"
#include "buff.h"
#include "daap.h"
#include "gzip.h"

extern daap_database kDefaultDatabase;
extern daap_adbs     kDefaultAdbs;
extern daap_apso     kDefaultApso;

static void
WriteFile (name, data, leng)
  char * name;
  char * data;
  int    leng;
{
  FILE * file;
  
  assert(name != NULL);
  assert(data != NULL);
  assert(leng >= 0);
  
  file = fopen(name, "w");
  fwrite(data, 1, leng, file);
  fclose(file);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Pool
#endif

http_pool *
http_pool_Create ()
{
  http_pool * pool;
  
  pool = (http_pool *)malloc(sizeof(http_pool));
  assert(pool != NULL);
  
  return pool;
}

void
http_pool_Init (http_pool * pool)
{
  int status;
  
  assert(pool != NULL);
  
  pool->len  = 0;
  pool->max  = 0;
  pool->pos  = 0;
  pool->list = NULL;

  pool->started = 0;
  pool->hault   = 0;

  FD_ZERO(&pool->fdset);
  pool->max_fd = 0;
  
  bzero(&pool->mutex, sizeof(pthread_mutex_t));
  bzero(&pool->thread, sizeof(pthread_t));
  
  status = pthread_mutex_init(&pool->mutex, NULL);
  assert(status == 0);
  
  http_pool_Resize(pool, 5);
}

int
http_pool_Resize (http_pool * pool, int size)
{
  int i = 0;
  int status = 0;
  
  assert(pool != NULL);
  assert(size >= 1);

  if (0 != (status = pthread_mutex_lock(&pool->mutex))) {
    fprintf(stderr, "http_pool_Resize.. failed to get mutex lock, %d\n", status);
    return -1;
  }
  
  pool->max += size;

  if (pool->list == NULL) 
    pool->list = (http_conn **)malloc(sizeof(http_conn *) * pool->max);
  else
    pool->list = (http_conn **)realloc(pool->list, sizeof(http_conn *) * pool->max);
  
  assert(pool->list != NULL);
  
  for (i = pool->max - size; i < pool->max; i++)
    pool->list[i] = NULL;

  fprintf(stderr, "http_pool_Resize.. increased by %d to %d\n", size, pool->max);

  if (0 != (status = pthread_mutex_unlock(&pool->mutex))) {
    fprintf(stderr, "http_pool_Resize.. failed to unlock mutex, %d\n", status);
    return -1;
  }
  
  return pool->max;
}

void
http_pool_Append (http_pool * pool, http_conn * conn)
{
  int i;
  int status;
  
  assert(pool != NULL);
  assert(conn != NULL);

  // check the present size of the pool to see if it is adequate
  // for the addition of another connection. we need to do this
  // before we grab the thread lock, because Resize() is also
  // going to try to get a lock ... and this would lead to a dead
  // lock ... which is no fun.
  if (pool->len == pool->max)
    http_pool_Resize(pool, pool->max * 1.5);
  
  if (0 != (status = pthread_mutex_lock(&pool->mutex))) {
    fprintf(stderr, "http_pool_Append.. failed to get mutex lock, %d\n", status);
    return;
  }

  for (i = 0; i < pool->max; i++)
    if (pool->list[i] == NULL)
      break;

  // now that we have found a happy home for the new connection, we
  // should place it there, increment the pool->len count, which denotes
  // the number of connections in the pool, and add the new connection
  // socket file descriptor to the master fd_set, fdset.
  if (i < pool->max) {
    pool->list[i] = conn;
    pool->len++;
    
    FD_SET(conn->sock, &pool->fdset);
    pool->max_fd = pool->max_fd < conn->sock ? conn->sock : pool->max_fd;
    
    // fprintf(stderr, "http_pool_Append.. appended conn at %d\n", i);
  }
  else
    fprintf(stderr, "http_pool_Append.. failed to append to pool\n");
  
  if (0 != (status = pthread_mutex_unlock(&pool->mutex))) {
    fprintf(stderr, "http_pool_Append.. failed to unlock mutex, %d\n", status);
    return;
  }
}

http_conn *
http_pool_Remove (pool, conn, lock)
  http_pool * pool;
  http_conn * conn;
  int         lock;
{
  int i = 0;
  int status = 0;
  
  assert(pool != NULL);
  assert(conn != NULL);
  
  if (lock == 0 && 0 != (status = pthread_mutex_lock(&pool->mutex))) {
    fprintf(stderr, "http_pool_Remove.. failed to get mutex lock, %d\n", status);
    return NULL;
  }
  
  for (i = 0; i < pool->max; i++)
    if (conn == pool->list[0]) {
      pool->list[i] = NULL;
      pool->len--;
      break;
    }
  
  if (lock == 0 && 0 != (status = pthread_mutex_unlock(&pool->mutex))) {
    fprintf(stderr, "http_pool_Remove.. failed to unlock mutex, %d\n", status);
    return NULL;
  }  
      
  return conn;
}

/**
void
http_pool_Start (http_pool * pool)
{
  int status = 0;
  
  assert(pool != NULL);
  
  if (pool->started != 0) {
    fprintf(stderr, "http_pool_Start.. thread is already started\n");
    return;
  }
  
  status = pthread_create(&pool->thread, NULL, http_pool_Loop, pool);
  assert(status == 0);
}

void
http_pool_Stop (http_pool * pool)
{
  assert(pool != NULL);
  
  if (pool->started != 1) {
    fprintf(stderr, "http_pool_Stop.. thread is not started\n");
    return;
  }
  
  pool->hault = 1;
}
**/

http_conn *
http_pool_GetNextReadable (pool)
  http_pool * pool;
{
  int i = 0;
  int status = 0;
  http_conn * conn;
  char buff;
  
  assert(pool != NULL);
  
  while (pool->hault == 0)
  {
    for (i = pool->pos; i < pool->max && pool->hault == 0; i++) {
      // if there are presently no connections in the pool,
      // then it is probably safe to sleep for a short while
      if (pool->len == 0) {
        usleep(10 * 10000);
        continue;
      }
      
      if (pool->list[i] == NULL)
        continue;
      
      conn = pool->list[i];
      status = recv(conn->sock, (char *)&buff, 1, MSG_PEEK);

      // check each pool->list[i]->sock for data availability
      // grab a pointer to it
      // remove it from the pool
      // return the pointer
      
      if (status == -1 && errno == EAGAIN) {
        usleep(10 * 10000); // one tenth of a second
        continue;
      }
      else if (status > 0) {
//      fprintf(stderr, "http_pool_GextNextReadable.. recv() reports data!\n");
        
        if (0 != (status = pthread_mutex_lock(&pool->mutex))) {
          fprintf(stderr, "http_pool_GetNextReadable.. failed to get mutex lock, %d\n", status);
          return NULL;
        }
        
        pool->list[i] = NULL;
        pool->len--;
        
        if (0 != (status = pthread_mutex_unlock(&pool->mutex))) {
          fprintf(stderr, "http_pool_GetNextReadable.. failed to unlock mutex, %d\n", status);
          return NULL;
        }
        
        return conn;
      }
      else {
        // fprintf(stderr, "http_pool_GextNextReadable.. recv() returned %d\n", status);
        
        if (0 != (status = pthread_mutex_lock(&pool->mutex))) {
          fprintf(stderr, "http_pool_GetNextReadable.. failed to get mutex lock, %d\n", status);
          return NULL;
        }
        
        pool->list[i] = NULL;
        pool->len--;
        
        if (0 != (status = pthread_mutex_unlock(&pool->mutex))) {
          fprintf(stderr, "http_pool_GetNextReadable.. failed to unlock mutex, %d\n", status);
          return NULL;
        }
        
        http_conn_Destroy(conn);
      }
    }
    
    pool->pos = 0;
  }
  
  return NULL;
}

void
http_pool_Destroy (http_pool * pool)
{
  int i;
  
  assert(pool != NULL);
  
  pthread_mutex_destroy(&pool->mutex);

  for (i = 0; i < pool->max; i++)
    if (pool->list[i] != NULL)
      free(pool->list[i]);
  
  free(pool->list);
  
  pool->len = 0;
  pool->max = 0;
  pool->pos = 0;
  pool->list = NULL;
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Connection
#endif

http_conn *
http_conn_Create ()
{
  http_conn * conn;
  
  conn = (http_conn *)malloc(sizeof(http_conn));
  assert(conn != NULL);
  
  return conn;
}

void
http_conn_Init (http_conn * conn)
{
  assert(conn != NULL);
  
  conn->sock        = 0;
  conn->saddr_len   = sizeof(struct sockaddr_in);
  conn->remote_port = 0;
  conn->local_port  = 0;  
//conn->readbuf_len = 0;
//conn->readbuf_max = 0;
  conn->readbuf     = buff_Create();
  
  buff_Init(conn->readbuf, 32 * 1000);
  
  bzero(&conn->remote_addr, 16);
  bzero(&conn->local_addr,  16);  
  bzero(&conn->saddr, sizeof(struct sockaddr_in));
}

void
http_conn_SetInfo (http_conn * conn, http_lstn * lstn)
{
  assert(conn != NULL);
  assert(lstn != NULL);
  
  conn->local_port = conn->saddr.sin_port;
  conn->remote_port = lstn->saddr.sin_port;
  
  inet_ntop(AF_INET, &conn->saddr.sin_addr, (char *)&conn->remote_addr, conn->saddr_len);
  inet_ntop(AF_INET, &lstn->saddr.sin_addr, (char *)&conn->local_addr,  lstn->saddr_len);
}

void
http_conn_SetNonBlock (conn)
  http_conn * conn;
{
  assert(conn != NULL);
  
  fcntl(conn->sock, F_SETFL, O_NONBLOCK);
}

/**
void
http_conn_Loop (http_conn * conn)
{
  
}
**/

void
http_conn_Destroy (http_conn * conn)
{
  assert(conn != NULL);
  
  if (conn->readbuf != NULL)
    buff_Destroy(conn->readbuf);

  fprintf(stderr, "http_conn_Destroy.. destroying connection [%s]\n", conn->remote_addr);
  
  conn->sock        = 0;
  conn->saddr_len   = 0;
  conn->remote_port = 0;
  conn->local_port  = 0;  
//conn->readbuf_len = 0;
//conn->readbuf_max = 0;
  conn->readbuf     = NULL;
  
  bzero(&conn->remote_addr, 16);
  bzero(&conn->local_addr,  16);  
  bzero(&conn->saddr, sizeof(struct sockaddr_in));
  
  free(conn);
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Listener
#endif

http_lstn *
http_lstn_Create ()
{
  http_lstn * lstn;
  
  lstn = (http_lstn *)malloc(sizeof(http_lstn));
  assert(lstn != NULL);
  
  return lstn;
}

void
http_lstn_Init (http_lstn * lstn)
{
  assert(lstn != NULL);
  
  lstn->sock        = 0;
  lstn->saddr_len   = sizeof(struct sockaddr_in);
  lstn->port        = 0;
  lstn->accept_cnt  = 0;
  lstn->pool        = NULL;
  lstn->started     = 0;
  lstn->hault       = 0;
  
  bzero(&lstn->saddr, sizeof(struct sockaddr_in));
  bzero(&lstn->thread, sizeof(pthread_t));
}

void
http_lstn_SetPort (http_lstn * lstn, int port)
{
  assert(lstn != NULL);
  assert(port >= 1);
  
  if (lstn->started != 0) {
    fprintf(stderr, "http_lstn_SetPort.. listener is already started\n");
    return;
  }
  
  lstn->port = port;
}

void
http_lstn_SetPool (http_lstn * lstn, http_pool * pool)
{
  assert(lstn != NULL);
  assert(pool != NULL);
  
  if (lstn->started != 0) {
    fprintf(stderr, "http_lstn_SetPool.. listener is already started\n");
    return;
  }
  
  lstn->pool = pool;
}

void
http_lstn_Start (http_lstn * lstn)
{
  int status = 0;
  
  assert(lstn != NULL);
  
  if (lstn->started != 0) {
    fprintf(stderr, "http_lstn_Start.. listener is already started\n");
    return;
  }
  
  if (lstn->pool == NULL) {
    fprintf(stderr, "http_lstn_Start.. pool not defined\n");
    return;
  }

  if (lstn->port <= 0) {
    fprintf(stderr, "http_lstn_Start.. port not defined\n");
    return;
  }
  
  status = pthread_create(&lstn->thread, NULL, http_lstn_Loop, lstn);
  assert(status == 0);
}

void
http_lstn_Stop (http_lstn * lstn)
{
  assert(lstn != NULL);
  
  if (lstn->started != 1) {
    fprintf(stderr, "http_lstn_Stop.. listener is not running\n");
    return;
  }
  
  lstn->hault = 1;
}

void *
http_lstn_Loop (void * arg)
{
  int status = 0;
  http_conn * conn;
  http_lstn * lstn = arg;
  
  assert(arg != NULL);
  
  fprintf(stderr, "http_lstn_Loop.. starting\n");
  lstn->started = 1;

  // try to create a new socket file descriptor. if it fails, reset the
  // handful of miscellaneous state flags, and return.
  if (0 >= (lstn->sock = socket(AF_INET, SOCK_STREAM, 0))) {
    fprintf(stderr, "http_lstn_Loop.. socket() failed, %d\n", lstn->sock);
    lstn->started = 0;
    lstn->hault   = 0;
    lstn->sock    = 0;
    return NULL;
  }
  
  lstn->saddr.sin_family = AF_INET;
  lstn->saddr.sin_addr.s_addr = htonl(INADDR_ANY);
  lstn->saddr.sin_port = htons(lstn->port);

  // try to bind the new socket. if it fails, close the socket, reset the 
  // handful of miscellaneous state flags, and return.
  if (0 != (status = bind(lstn->sock, (struct sockaddr *)&lstn->saddr, sizeof(struct sockaddr)))) {
    fprintf(stderr, "http_lstn_Loop.. bind() failed, %d\n", status);
    close(lstn->sock);
    lstn->started = 0;
    lstn->hault   = 0;
    lstn->sock    = 0;
    return NULL;
  }

  // try to start the friendly bound, new socket, listening. if it fails,
  // close the socket, reset the handful of miscellaneous state flags, and
  // return.
  if (0 != (status = listen(lstn->sock, 16))) {
    fprintf(stderr, "http_lstn_Loop.. listen() failed, %d\n", status);
    close(lstn->sock);
    lstn->started = 0;
    lstn->hault   = 0;
    lstn->sock    = 0;
    return NULL;
  }

  // create a new http_conn. this is used to hold, among other things,
  // the socket file descriptor, and the sockaddr_in structure. we
  // create a new http_conn each time, after accepting a new connection
  conn = http_conn_Create();
  assert(conn != NULL);
  http_conn_Init(conn);
  
  while (lstn->hault == 0)
  {
    if (0 >= (conn->sock = accept(lstn->sock, (struct sockaddr *)&conn->saddr, (int *)&conn->saddr_len))) {
      fprintf(stderr, "http_lstn_Loop.. accept failed, %d\n", conn->sock);
      continue;
    }

    fprintf(stderr, "http_lstn_Loop.. accepted new connection [sock=%d]\n", conn->sock);

    lstn->accept_cnt++;
    
    // set the non-blocking flag on the new connection's socket, set
    // the local and remote address and port information, and add 
    // it all to the designated connection pool
    http_conn_SetNonBlock(conn);
    http_conn_SetInfo(conn, lstn);
    http_pool_Append(lstn->pool, conn);

    fprintf(stderr, "http_lstn_Loop.. local  - %s:%d\n", (char *)&conn->local_addr,  conn->local_port);
    fprintf(stderr, "http_lstn_Loop.. remote - %s:%d\n", (char *)&conn->remote_addr, conn->remote_port);
    
    // having now used the previously created http_conn for a new connection,
    // we need to make a new one in preparation for the next connection.
    conn = http_conn_Create();
    assert(conn != NULL);
    http_conn_Init(conn);
  }

  // having exited the loop, we should close tho socket, and clean
  // everything back up to the point it was at prior to entering
  // this function ... as close as possible, anyway.
  close(lstn->sock);
  lstn->started = 0;
  lstn->hault   = 0;
  lstn->sock    = 0;
  
  fprintf(stderr, "http_lstn_Loop.. stopping\n");
  
  return NULL;
}

void
http_lstn_Destroy (http_lstn * lstn)
{
  
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Surf
#endif

//
// drawing on my present fixation with four-letter "types" (ie, lstn, conn, 
// pool, etc), "worker" or "process" just weren't going to cut it. although,
// "proc" could suffice. but, as they're is a much cooler, and more applicable
// definition for the word "serf", i am going to use that, instead. the
// definition, as stolen from dictionary.com, follows;
//
//   A member of the lowest feudal class, attached to 
//   the land owned by a lord and required to perform 
//   labor in return for certain legal or customary rights.
//
// so as to clarify the purpose of this "class", "process" would probably be
// a much better description. as such, just think about it as that. more to
// the point, this class represents the worker ... it waits for incoming data
// and processes it ... doing all of the "dirty work". thus serf is such an
// appropriate name - it's just somewhat non-traditional
//

http_serf *
http_serf_Create ()
{
  http_serf * serf;
  
  serf = (http_serf *)malloc(sizeof(http_serf));
  assert(serf != NULL);
  
  return serf;
}

void
http_serf_Init (serf, pool)
  http_serf * serf;
  http_pool * pool;
{
  assert(serf != NULL);
  assert(pool != NULL);
  
  serf->request_cnt = 0;
  serf->started     = 0;
  serf->hault       = 0;
  serf->pool        = pool;
  
  bzero(&serf->thread, sizeof(pthread_t));
}

void
http_serf_Start (serf)
  http_serf * serf;
{
  int status = 0;
  
  assert(serf != NULL);
  
  if (serf->started != 0) {
    fprintf(stderr, "http_serf_Start.. serf already hard at work\n");
    return;
  }
  
  if (serf->pool == NULL) {
    fprintf(stderr, "http_serf_Pool.. pool not defined\n");
    return;
  }
  
  status = pthread_create(&serf->thread, NULL, http_serf_Loop, serf);
  assert(status == 0);
}

void
http_serf_Stop (serf)
  http_serf * serf;
{
  assert(serf != NULL);
  
  if (serf->started != 1) {
    fprintf(stderr, "http_serf_Stop.. serf is already not working\n");
    return;
  }
  
  serf->hault = 1;
}

/**
static int
http_serf_ReadLine (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  assert(serf != NULL);
  assert(conn != NULL);
  
  int i = 0;
  buff_t * buff = conn->readbuf;
  
  for (i = 0; i < conn->readbuf_len - 1; i++)
    if (*(conn->readbuf + i) == '\r' && *(conn->readbuf + i + 1) == '\n')
      return i + 2;
  
  return -1;
}
**/

void *
http_serf_Loop (arg)
  void * arg;
{
  int status = 0;
  int result = 0;
  http_serf * serf = arg;
  http_conn * conn;
  buff_t    * buff;
  char data[500];
  int left;
  int size;
  
  assert(arg != NULL);

  fprintf(stderr, "http_serf_Loop.. starting\n");
  serf->started = 1;
  
  while (serf->hault == 0)
  {
    if (NULL == (conn = http_pool_GetNextReadable(serf->pool))) {
      fprintf(stderr, "http_serf_Loop.. GetNextReadable() returned null\n");
      continue;
    }

    // fprintf(stderr, "http_serf_Loop.. data to be read from: %s\n", conn->remote_addr);

    // as the size of our little temporary data buff is presently 500 bytes,
    // we don't want to try to read any more than that, at once, from the 
    // socket. however, it is a distinct possibility that the buff_t into
    // which we will be placing this data lacks sufficient space to hold
    // another 500 bytes. thus, we must temper the rate at which we read
    // data to ensure we do not exceed the buff_t's capacity.
    //
    buff = conn->readbuf;
    left = buff_Remaining(buff);
    size = 500 > left ? left : 500;
    // fprintf(stderr, "http_serf_Loop.. [1] will try to read %d bytes\n", size);
      
    // try to read data from the socket, up to the remaining capacity of this
    // connection1s read buffer, stopping then, or when there is no longer
    // any data to be read - and as this is non-blocking, we'll just move right
    // along.
    //
    while (0 < (result = read(conn->sock, &data, size))) {
//    fprintf(stderr, "http_serf_Loop.. [2] read %d bytes\n", result);
      
      buff_Append(buff, &data, result);
      
      left = buff_Remaining(buff);
      size = 500 > left ? left : 500;
    }

//  fprintf(stderr, "http_serf_Loop.. [3] done reading from socket\n");
    
    if (result == -1 && errno == EAGAIN) {
      status = http_serf_Process(serf, conn);
      
      if (status == 0)
        http_pool_Append(serf->pool, conn);
      else
        http_conn_Destroy(conn);
      
      continue;
    }
    else if (result == -1) {
      fprintf(stderr, "http_serf_Loop.. read() failed with, %d\n", result);
      http_conn_Destroy(conn);
      continue;
    }
    else if (result == 0) {
      fprintf(stderr, "http_serf_Loop.. read() reported EOF\n");
      http_conn_Destroy(conn);
      continue;
    }
  }
  
  serf->started = 0;
  serf->hault   = 0;
  
  fprintf(stderr, "http_serf_Loop.. stopping\n");
  
  return NULL;
}

int
http_serf_Process (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  int result = 0;
  buff_t * buff;
  int linelen = 0;
  char data[1000];
  int i;
  int mark_s;
  int mark_q;
  int mark_x;
  
  assert(serf != NULL);
  assert(conn != NULL);

  buff = conn->readbuf;
  
  while (-1 != (linelen = buff_ReadLine(buff))) {
    // fprintf(stderr, "http_serf_Process.. buff_ReadLine() linelen = %d\n", linelen);
    
    if (linelen == 2) {
      buff_Substring(buff, NULL, 0, linelen);
      
      // reached the end of the HTTP header section
      // now is the time to actually process the request

      http_serf_Execute(serf, conn);
      
      if (conn->query != NULL) {
        http_query_Destroy(conn->query);
        conn->query = NULL;
      }
      
      break;
    }
    else if (linelen > 1000) {
      fprintf(stderr, "http_serf_Process.. line too long (%d); flushing\n", linelen);
      buff_Substring(buff, NULL, 0, linelen);
      continue;
    }

    bzero(&data, 1000);

    result = buff_Substring(buff, (char *)&data, 0, linelen);
    fprintf(stderr, "http_serf_Process.. %s", (char *)&data);
    
//  fprintf(stdout, "%s DATA[%d]: ", (char *)&conn->remote_addr, result - 2);
//  fwrite(&data, sizeof(char), result - 2, stdout);
//  fprintf(stdout, "\n");
//  fflush(stdout);
    
    if (0 == strncmp((char *)&data, "GET /", 5)) {
      conn->query = http_query_Create();
      http_query_Init(conn->query);

      i = 0;
      mark_s = -1;        // space
      mark_q = -1;        // question mark
      mark_x = -1;        // end of file path
      
      for (i = 4; i < linelen; i++) {
        if (*(data + i) == ' ') {
          mark_s = i;
          break;
        }
      }

      for (i = 4; i < linelen; i++) {
        if (*(data + i) == '?') {
          mark_q = i;
          break;
        }
      }

      mark_x = mark_q == -1 ? mark_s : mark_q;
      
      http_query_SetFilePath(conn->query, data + 4, mark_x - 4);
      
      if (mark_q != -1)
        http_query_SetCGIQuery(conn->query, data + mark_q, mark_s - mark_q);

      continue;
    }
    // an http header value of the form, Name: Value\r\n
    else {
      int i_colon = 0;
      int i       = 0;
      
      for (i = 0; i < linelen; i++)
        if (data[i] == ':')
          break;

      if (i >= linelen - 1)
        continue;
      
      i_colon = i;
      data[i_colon] = 0;
      
      if (0 == strncmp(data, "Range", 5)) {
        unsigned char bytes[20];
        bzero((unsigned char *)&bytes, 20);
        memcpy(&bytes, data + 7, linelen - 9);
        conn->query->brange = atoi((char *)&bytes);
        fprintf(stderr, "http_serf_Process.. byte range = %d\n", conn->query->brange);
      }

    }
  }
  
  return 0;
}

void
http_serf_Execute (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  assert(serf != NULL);
  assert(conn != NULL);

  if (0 == strncmp((char *)conn->query->file_path, "/server-info", 12))
    http_serf_ExecuteServerInfoRequest(serf, conn);
  else if (0 == strncmp((char *)conn->query->file_path, "/content-codes", 14))
    http_serf_ExecuteContentCodesRequest(serf, conn);
  else if (0 == strncmp((char *)conn->query->file_path, "/login", 6))
    http_serf_ExecuteLoginRequest(serf, conn);
  else if (0 == strncmp((char *)conn->query->file_path, "/update", 7)) {
    if (conn->query->cgi_len < 35)
      http_serf_ExecuteUpdateRequest(serf, conn);
    else {
      
    }
  }
  else if (0 == strncmp((char *)conn->query->file_path, "/databases", 10))
    http_serf_ExecuteDatabasesRequest(serf, conn);
//else
//  http_serf_ExecuteFileNotFound(serf, conn);
}

void
http_serf_ExecuteServerInfoRequest (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  buff_t * buff;
  char temp[30];
  int templen = 0;
  
  daap_msrv * msrv;
  buff_t * buff2;
  buff_t * buff9;
  
  assert(serf != NULL);
  assert(conn != NULL);
  
  buff = buff_Create();
  assert(buff != NULL);
  buff_Init(buff, 1000);
  
  msrv = msrv_Create();
  msrv_Init(msrv);
  buff2 = msrv_Serialize(msrv);
  buff9 = gzip_Compress(buff2->data, buff2->leng);
  
  buff_Append(buff, "HTTP/1.1 200 OK\r\n",                          17);
  buff_Append(buff, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",      37);
  buff_Append(buff, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",       36);
  buff_Append(buff, "Content-Type: application/x-dmap-tagged\r\n",  41);
  buff_Append(buff, "Content-Length: ",                             16);
  templen = sprintf((char *)&temp, "%d", buff9->leng);
  buff_Append(buff, &temp, templen);
  buff_Append(buff, "\r\nContent-Encoding: gzip",                   24);
  buff_Append(buff, "\r\n\r\n",                                      4);

  write(conn->sock, buff->data, buff->leng);
  write(conn->sock, buff9->data, buff9->leng);
  
  WriteFile("./01_server-info.head", buff->data,  buff->leng);
  WriteFile("./01_server-info.body", buff2->data, buff2->leng);
  
  buff_Destroy(buff);
  buff_Destroy(buff2);
  msrv_Destroy(msrv);
  buff_Destroy(buff9);
}

void
http_serf_ExecuteContentCodesRequest (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  buff_t * buff;
  char temp[30];
  int templen = 0;
  
  daap_mccr * mccr;
  buff_t * buff2;
  buff_t * buff9;
  
  assert(serf != NULL);
  assert(conn != NULL);
  
  buff = buff_Create();
  assert(buff != NULL);
  buff_Init(buff, 10000);
  
  mccr = mccr_Create();
  mccr_Init(mccr);
  buff2 = mccr_Serialize(mccr);
  buff9 = gzip_Compress(buff2->data, buff2->leng);
  
  buff_Append(buff, "HTTP/1.2 200 OK\r\n",                          17);
  buff_Append(buff, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",      37);
  buff_Append(buff, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",       36);
  buff_Append(buff, "Content-Type: application/x-dmap-tagged\r\n",  41);
  buff_Append(buff, "Content-Encoding: gzip\r\n",                   24);
  buff_Append(buff, "Content-Length: ",                             16);
  templen = sprintf((char *)&temp, "%d", buff9->leng);
  buff_Append(buff, &temp, templen);
  buff_Append(buff, "\r\n\r\n",                                      4);
  
  // buff_Append(buff, buff2->data, buff2->leng);
  
  write(conn->sock, buff->data, buff->leng);
  write(conn->sock, buff9->data, buff9->leng);
  
  WriteFile("./02_content-codes.head", buff->data,  buff->leng);
  WriteFile("./02_content-codes.body", buff2->data, buff2->leng);
  
  buff_Destroy(buff);
  buff_Destroy(buff2);
  mccr_Destroy(mccr);
  buff_Destroy(buff9);
}

void
http_serf_ExecuteLoginRequest (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  buff_t * buff;
  char temp[30];
  int templen = 0;
  
  daap_mlog * mlog;
  buff_t * buff2;
  buff_t * buff9;
  
  assert(serf != NULL);
  assert(conn != NULL);
  
  buff = buff_Create();
  assert(buff != NULL);
  buff_Init(buff, 10000);
  
  mlog = mlog_Create();
  mlog_Init(mlog);
  buff2 = mlog_Serialize(mlog);
  buff9 = gzip_Compress(buff2->data, buff2->leng);
  
  buff_Append(buff, "HTTP/1.2 200 OK\r\n",                          17);
  buff_Append(buff, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",      37);
  buff_Append(buff, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",       36);
  buff_Append(buff, "Content-Type: application/x-dmap-tagged\r\n",  41);
  buff_Append(buff, "Content-Encoding: gzip\r\n",                   24);
  buff_Append(buff, "Content-Length: ",                             16);
  templen = sprintf((char *)&temp, "%d", buff9->leng);
  buff_Append(buff, &temp, templen);
  buff_Append(buff, "\r\n\r\n",                                      4);
  
  // buff_Append(buff, buff2->data, buff2->leng);
  
  write(conn->sock, buff->data, buff->leng);
  write(conn->sock, buff9->data, buff9->leng);
  
  WriteFile("./03_login.head", buff->data,  buff->leng);
  WriteFile("./03_login.body", buff2->data, buff2->leng);
  
  buff_Destroy(buff);
  buff_Destroy(buff2);
  mlog_Destroy(mlog);
  buff_Destroy(buff9);
}

void
http_serf_ExecuteUpdateRequest (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  buff_t * buff;
  char temp[30];
  int templen = 0;
  
  daap_mupd * mupd;
  buff_t * buff2;
  buff_t * buff9;
  
  assert(serf != NULL);
  assert(conn != NULL);
  
  buff = buff_Create();
  assert(buff != NULL);
  buff_Init(buff, 10000);

  mupd = mupd_Create();
  mupd_Init(mupd);
  buff2 = mupd_Serialize(mupd);
  buff9 = gzip_Compress(buff2->data, buff2->leng);
  
  buff_Append(buff, "HTTP/1.2 200 OK\r\n",                          17);
  buff_Append(buff, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",      37);
  buff_Append(buff, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",       36);
  buff_Append(buff, "Content-Type: application/x-dmap-tagged\r\n",  41);
  buff_Append(buff, "Content-Encoding: gzip\r\n",                   24);
  buff_Append(buff, "Content-Length: ",                             16);
  templen = sprintf((char *)&temp, "%d", buff9->leng);
  buff_Append(buff, &temp, templen);
  buff_Append(buff, "\r\n\r\n",                                      4);
  
  // buff_Append(buff, buff2->data, buff2->leng);
  
  write(conn->sock, buff->data,  buff->leng);
  write(conn->sock, buff9->data, buff9->leng);
  
  WriteFile("./04_update.head", buff->data,  buff->leng);
  WriteFile("./04_update.body", buff2->data, buff2->leng);
  
  buff_Destroy(buff);
  buff_Destroy(buff2);
  mupd_Destroy(mupd);
  buff_Destroy(buff9);
}

void
http_serf_ExecuteDatabasesRequest (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
//int status = 0;
  buff_t * buff;
  char temp[30];
  int templen = 0;
  
  buff_t * buff2 = NULL;
  buff_t * buff9 = NULL;
  
  daap_avdb * avdb = NULL;
  daap_adbs * adbs = (daap_adbs *)&kDefaultAdbs;
  daap_aply * aply = NULL;
  daap_apso * apso = (daap_apso *)&kDefaultApso;
  
  assert(serf != NULL);
  assert(conn != NULL);

  buff = buff_Create();
  assert(buff != NULL);
  buff_Init(buff, 1000000);
  
  if (0 == strncmp((char *)conn->query->file_path, "/databases/32/containers/1/items", 32)) {
    buff2 = apso_Serialize(apso);
  }
  else if (0 == strncmp((char *)conn->query->file_path, "/databases/32/items/", 20)) {
    http_serf_ExecuteStreamSong(serf, conn);
    buff_Destroy(buff);
    return;
  }
  else if (0 == strncmp((char *)conn->query->file_path, "/databases/32/containers", 24)) {
    aply = aply_Create();
    aply_Init(aply);
    buff2 = aply_Serialize(aply);
  }
  else if (0 == strncmp((char *)conn->query->file_path, "/databases/32/items", 19)) {
    buff2 = adbs_Serialize(adbs);
  }
  else if (0 == strncmp((char *)conn->query->file_path, "/databases", 10)) {
    avdb = avdb_Create();
    avdb_Init(avdb);
    buff2 = avdb_Serialize(avdb);
  }
  else {
    fprintf(stderr, "http_serf_ExecuteDatabasesRequest.. invalid request, '%s'\n", (char *)conn->query->file_path);
    return;
  }

  buff9 = gzip_Compress(buff2->data, buff2->leng);
  
  buff_Append(buff, "HTTP/1.2 200 OK\r\n",                          17);
  buff_Append(buff, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",      37);
  buff_Append(buff, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",       36);
  buff_Append(buff, "Content-Type: application/x-dmap-tagged\r\n",  41);
  buff_Append(buff, "Content-Encoding: gzip\r\n",                   24);
  buff_Append(buff, "Content-Length: ",                             16);
  templen = sprintf((char *)&temp, "%d", buff9->leng);
  buff_Append(buff, &temp, templen);
  buff_Append(buff, "\r\n\r\n",                                      4);

  write(conn->sock, buff->data,  buff->leng);
  write(conn->sock, buff9->data, buff9->leng);

  if (apso != NULL) {
    // apso_Destroy(apso);
    WriteFile("./08_databases.head", buff->data,  buff->leng);
    WriteFile("./08_databases.body", buff2->data, buff2->leng);
  }

  if (aply != NULL) {
    aply_Destroy(aply);
    WriteFile("./07_databases.head", buff->data,  buff->leng);
    WriteFile("./07_databases.body", buff2->data, buff2->leng);
  }
  
  if (adbs != NULL) {
    // adbs_Destroy(adbs);
    WriteFile("./06_databases.head", buff->data,  buff->leng);
    WriteFile("./06_databases.body", buff2->data, buff2->leng);
  }
  
  if (avdb != NULL) {
    avdb_Destroy(avdb);
    WriteFile("./05_databases.head", buff->data,  buff->leng);
    WriteFile("./05_databases.body", buff2->data, buff2->leng);
  }
  
  buff_Destroy(buff);
  buff_Destroy(buff2);
  buff_Destroy(buff9);
}

void
http_serf_ExecuteStreamSong (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  daap_song * song;
  FILE * file;
  
  int i_start = 20;
  int i_stop  = 0;
  int i_numb  = 0;
  int i;
  
  unsigned char data[10000];
  unsigned int rleng = 0;
  unsigned int wleng = 0;
  
  assert(serf != NULL);
  assert(conn != NULL);

  // Range: bytes=12345
  
  fprintf(stderr, "http_serf_ExecuteStreamSong.. %s\n", conn->query->file_path);
  
  // /databases/32/items/1.mp3
  for (i = i_start; i < conn->query->file_len && i_stop == 0; i++)
    if (conn->query->file_path[i] < 0x30 || conn->query->file_path[i] > 0x39)
      i_stop = i;

  conn->query->file_path[i_stop] = 0;
  i_numb = atoi(conn->query->file_path + i_start);
  
  fprintf(stderr, "http_serf_ExecuteStreamSong.. looking for entry %d\n", i_numb);
  
  if (kDefaultDatabase.len < i_numb) {
    fprintf(stderr, "http_serf_ExecuteStreamSong.. database [%d] not big enough for %d position\n",
            kDefaultDatabase.len, i_numb);
    http_serf_ExecuteFileNotFound(serf, conn);
    return;
  }
  
  if (kDefaultDatabase.songs[i_numb-1] == NULL) {
    fprintf(stderr, "http_serf_ExecuteStreamSong.. database entry at %d is NULL\n", i_numb);
    http_serf_ExecuteFileNotFound(serf, conn);
    return;
  }
  
  song = kDefaultDatabase.songs[i_numb-1];
  
  if (song->file_path == NULL) {
    fprintf(stderr, "http_serf_ExecuteStreamSong.. file path is NULL at %d\n", i_numb);
    http_serf_ExecuteFileNotFound(serf, conn);
    return;
  }

  fprintf(stderr, "http_serf_ExecuteStreamSong.. %s\n", song->file_path);
  
  file = fopen(song->file_path, "r");
  assert(file != NULL);

  if (conn->query->brange > 0)
    rleng = fseek(file, conn->query->brange, SEEK_SET);
  
  {
    char head[500];
    bzero((char *)&head, 500);
    
    memcpy((char *)&head +   0, "HTTP/1.2 200 OK\r\n",                           17);
    memcpy((char *)&head +  17, "Date: Tue, 27 Jan 2004 06:48:41 GMT\r\n",       37);
    memcpy((char *)&head +  54, "DAAP-Server: iTunes/4.2 (Mac OS X)\r\n",        36);
    memcpy((char *)&head +  90, "Content-Type: application/x-dmap-tagged\r\n",   41);
    memcpy((char *)&head + 131, "Accept-Ranges: bytes\r\n",                      22);
    memcpy((char *)&head + 153, "Content-Length: ",                              16);
    
    sprintf((char *)&head + 169, "%d\r\n\r\n", (song->assz.valu - rleng));
    
    write(conn->sock, &head, strlen((char *)&head));
  }
  
  rleng = fread(&data, 1, 10000, file);
  
  while (rleng != 0)
  {
    wleng = write(conn->sock, &data, rleng);
    
    if (wleng == -1 && errno == EAGAIN) {
      sleep(1);
      continue;
    }
    else if (wleng == -1) {
      break;
    }
    else if (wleng < rleng) {
      memcpy(data, data + wleng, rleng - wleng);
      rleng -= wleng;
      continue;
    }
    
    rleng = fread(&data, 1, 10000, file);
  }

  fprintf(stderr, "http_serf_ExecuteStreamSong.. finished streaming\n");
  
  // close client connection after removing from the http_pool
  close(conn->sock);
  
  fclose(file);
}

void
http_serf_ExecuteFileNotFound (serf, conn)
  http_serf * serf;
  http_conn * conn;
{
  
}

void
http_serf_Destroy (serf)
  http_serf * serf;
{
  
}

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Query
#endif

http_query *
http_query_Create ()
{
  http_query * query;
  
  query = (http_query *)malloc(sizeof(http_query));
  assert(query != NULL);
  
  return query;
}

void
http_query_Init (query)
  http_query * query;
{
  assert(query != NULL);
  
  bzero(query, sizeof(http_query));
  
  query->arg_len  = 0;
  query->arg_max  = 5;
  query->file_len = 0;
  query->cgi_len  = 0;
  query->brange   = 0;
  
  query->file_path = NULL;
  query->cgi_query = NULL;
  
  query->argv = (http_cgiv **)malloc(sizeof(http_cgiv *) * query->arg_max);
  assert(query->argv != NULL);
}

void
http_query_SetFilePath (query, file, leng)
  http_query * query;
  char * file;
  int leng;
{
  assert(query != NULL);
  assert(file  != NULL);
  assert(leng  >= 0);
  
  if (query->file_path == NULL)
    query->file_path = (char *)malloc(leng + 1);
  else
    query->file_path = (char *)realloc(query->file_path, leng + 1);
  
  assert(query->file_path != NULL);
  bzero(query->file_path, leng + 1);
  memcpy(query->file_path, file, leng);
  
  query->file_len = leng;
  
  fprintf(stdout, "http_query_SetFilePath.. [%03d]: '", query->file_len);
  fwrite(query->file_path, 1, query->file_len, stdout);
//fflush(stdout);
  fprintf(stdout, "'\n");
}

void
http_query_SetCGIQuery (query, cgistr, leng)
  http_query * query;
  char * cgistr;
  int leng;
{
  assert(query  != NULL);
  assert(cgistr != NULL);
  assert(leng   >= 0);
  
  if (query->cgi_query == NULL)
    query->cgi_query = (char *)malloc(leng + 1);
  else
    query->cgi_query = (char *)realloc(query->cgi_query, leng + 1);
  
  assert(query->cgi_query != NULL);
  bzero(query->cgi_query, leng + 1);
  memcpy(query->cgi_query, cgistr, leng);
  
  query->cgi_len = leng;
  
  // fprintf(stdout, "http_query_SetCGIQuery.. [%d]: '", query->cgi_len);
  // fwrite(query->cgi_query, 1, query->cgi_len, stdout);
  // fflush(stdout);
  // fprintf(stdout, "'\n");
}

void
http_query_AddArg (query, name, namelen, valu, valulen)
  http_query * query;
  char * name;
  int namelen;
  char * valu;
  int valulen;
{
  http_cgiv * cgiv;
  
  assert(query != NULL);
  assert(name  != NULL);
  assert(valu  != NULL);
  assert(namelen > 0);
  assert(valulen > 0);
  
  cgiv = (http_cgiv *)malloc(sizeof(http_cgiv));
  assert(cgiv != NULL);
  
  cgiv->name = (char *)malloc(sizeof(char) * namelen);
  cgiv->valu = (char *)malloc(sizeof(char) * valulen);
  
  memcpy(cgiv->name, name, namelen);
  memcpy(cgiv->valu, valu, valulen);

  if (query->arg_len == query->arg_max) {
    query->arg_max *= 1.5;
    query->argv = (http_cgiv **)realloc(query->argv, sizeof(http_cgiv *) * query->arg_max);
    assert(query->argv != NULL);
  }
  
  query->argv[query->arg_len++] = cgiv;
}

void
http_query_Destroy (query)
  http_query * query;
{
  int i = 0;
  
  assert(query != NULL);
  
  if (query->file_path != NULL)
    free(query->file_path);
  
  if (query->cgi_query != NULL)
    free(query->cgi_query);

  if (query->argv != NULL) {
    for (i = 0; i < query->arg_len; i++) {
      if (query->argv[i] != NULL) {
        free(query->argv[i]);
        query->argv[i] = NULL;
      }
    }
    
    free(query->argv);
    query->argv = NULL;
  }

  query->file_path = NULL;
  query->cgi_query = NULL;
  
  query->arg_len  = 0;
  query->file_len = 0;
  query->cgi_len  = 0;
  
  free(query);
}
