#ifndef __HTTP_H
#define __HTTP_H

#include <time.h>             // time()
#include <sys/types.h>        //
#include <sys/socket.h>       //
#include <sys/select.h>       // select()
#include <netinet/in.h>       //
#include <unistd.h>           //
#include <pthread.h>          //
#include <errno.h>            // 

#include "buff.h"

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Struct Definitions
#endif

extern int errno;

typedef struct                // http cgi parameters ... name=value pairs, that is
{                             //
  int name_len;               // length of the name field
  int valu_len;               // length of the value field
  char * name;                // name of the parameter
  char * valu;                // value of the parameter
} http_cgiv;

typedef struct                // represents an entire parsed http query
{                             //
  int arg_len;                // number of http_cgiv's in argv
  int arg_max;                // malloc'ed size of argv array
  int file_len;               // length of the file_path string
  int cgi_len;                // length of the cgi_query string
  int brange;                 // byte range value in http request
  char * file_path;           // http GET request, file name and path
  char * cgi_query;           // cgi query, starting with a '?'
  http_cgiv ** argv;          // cgi parameters ... name=value pairs, that is
} http_query;

typedef struct
{
  int sock;                   // socket for http connection
  char remote_addr[16];       // remote client's ip address
  int remote_port;            // remote client's port number
  char local_addr[16];        // local ip address to which client is connected
  int local_port;             // local port number to which client is connected
  time_t keep_alive;          // time at which connection will expire
  struct sockaddr_in saddr;   // socket address data structure
  int saddr_len;              // length of aforementioned data structure
  buff_t * readbuf;           // thread safe read buffer
  http_query * query;         // structure stores all query related data
} http_conn;

typedef struct
{
  pthread_mutex_t mutex;      // thread synchronization mutex
  pthread_t thread;           // thread.. waits on select for conns
  fd_set fdset;               // set of sockets for use with select
  int max_fd;                 // maximum value of all socket file descriptors
  int len;                    // number of connections in pool
  int max;                    // malloc'ed size of connection pool
  int pos;                    // the iterator's position in the pool
  int started;                // whether the thread is running
  int hault;                  // flag to denote that thread should stop
  http_conn ** list;          // list of http_conn pointers
} http_pool;

typedef struct                // holds all of the values pertaining to a listener
{                             //
  int sock;                   // file descriptor for listening socket
  int port;                   // port on which to bind the listener
  int accept_cnt;             // number of connections accepted
  int started;                // whether the listener is running
  int hault;                  // a flag to denote that the loop should cease
  struct sockaddr_in saddr;   // socket address data structure
  int saddr_len;              // length of aforementioned data structure
  pthread_t thread;           // thread
  http_pool * pool;           // the pool to which connections should be added
} http_lstn;

typedef struct                // holds all values necessary for a serf thread
{                             //
  pthread_t thread;           // the thread for this worker process dude
  http_pool * pool;           // the pool on which this thread works
  int request_cnt;            // the number of requests this thread has handled
  int started;                // a flag denoting whether this thread has been started
  int hault;                  // a flag to signal that this thrtead should cease
} http_serf;

typedef struct
{
  pthread_t thread;
} http_thread;

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Pool
#endif

http_pool * http_pool_Create();
void http_pool_Init (http_pool *);
int http_pool_Resize (http_pool *, int size);
void http_pool_Append (http_pool *, http_conn * conn);
http_conn * http_pool_Remove (http_pool *, http_conn * conn, int lock);
// void http_pool_Start (http_pool *);
// void http_pool_Stop (http_pool *);
http_conn * http_pool_GetNextReadable (http_pool *);
// void * http_pool_Loop (void * arg);
void http_pool_Destroy (http_pool *);

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Connection
#endif

http_conn * http_conn_Create();
void http_conn_Init (http_conn *);
void http_conn_SetInfo (http_conn *, http_lstn * lstn);
// void http_conn_Loop (http_conn *);
void http_conn_Destroy (http_conn *);

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Listener
#endif

http_lstn * http_lstn_Create();
void http_lstn_Init (http_lstn *);
void http_lstn_SetPort (http_lstn *, int port);
void http_lstn_SetPool (http_lstn *, http_pool * pool);
void http_lstn_Start (http_lstn *);
void http_lstn_Stop (http_lstn *);
void * http_lstn_Loop (void * arg);
void http_lstn_Destroy (http_lstn *);

#if COMPILER_LIKES_PRAGMA_MARK
#pragma mark ***** HTTP Serf
#endif

http_serf * http_serf_Create();
void http_serf_Init (http_serf *, http_pool *);
void http_serf_Start (http_serf *);
void http_serf_Stop (http_serf *);
void * http_serf_Loop (void * arg);
int  http_serf_Process (http_serf *, http_conn * conn);
void http_serf_Execute (http_serf *, http_conn * conn);
void http_serf_ExecuteServerInfoRequest (http_serf *, http_conn *);
void http_serf_ExecuteContentCodesRequest (http_serf *, http_conn *);
void http_serf_ExecuteLoginRequest (http_serf *, http_conn *);
void http_serf_ExecuteUpdateRequest (http_serf *, http_conn *);
void http_serf_ExecuteDatabasesRequest (http_serf *, http_conn *);
void http_serf_ExecuteFileNotFound (http_serf *, http_conn *);
void http_serf_ExecuteStreamSong (http_serf *, http_conn *);
void http_serf_Destroy (http_serf *);

http_query * http_query_Create ();
void http_query_Init (http_query *);
void http_query_SetFilePath (http_query *, char * file, int leng);
void http_query_SetCGIQuery (http_query *, char * cgistr, int leng);
void http_query_AddArg (http_query *, char * name, int namelen, char * valu, int valulen);
void http_query_Destroy (http_query *);

#endif /* __HTTP_H */
